# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Strategy classes for aiohttp."""

from logging import getLogger

from ..._vendors.wrapt import FunctionWrapper
from .framework import FrameworkStrategy
from .import_hook import ImportHookStrategy

LOGGER = getLogger(__name__)


def get_aiohttp_middleware():
    """Return the aiohttp middleware to be installed.

    In aiohttp 2.2, middlewares are factory functions that return request
    handlers. In aiohttp 2.3, middleware factories (aka old-style middlewares)
    are deprecated in favor of simpler middleware functions, decorated with
    aiohttp.web.middleware. This function returns a suited middleware,
    depending on aiohttp version.
    """
    from ..middlewares.aiohttp_middleware import AioHTTPMiddleware

    try:
        from aiohttp import web
    except ImportError:
        LOGGER.debug("Cannot import aiohttp.web", exc_info=True)
        web = None
    if hasattr(web, "middleware"):
        LOGGER.debug("New style middlewares are supported")
        return AioHTTPMiddleware.handle
    LOGGER.debug("New style middlewares are not supported")
    return AioHTTPMiddleware.factory


class AioHTTPInstallStrategy(ImportHookStrategy):
    """Wrap aiohttp.web::Application.freeze to install a custom middleware."""

    def import_hook_callback(self, original):
        def wrapper(wrapped, instance, args, kwargs):
            middleware = get_aiohttp_middleware()
            if middleware in instance._middlewares:
                # Prevents inserting the middleware twice.
                LOGGER.warning("Custom middleware already set, skipped")
            else:
                LOGGER.info("Injecting custom middleware")
                instance._middlewares.insert(0, middleware)
            return wrapped(*args, **kwargs)

        return FunctionWrapper(original, wrapper)


class AioHTTPHookStrategy(FrameworkStrategy):
    """Hook aiohttp custom middleware."""

    MODULE_NAME = "sqreen.instrumentation.middlewares.aiohttp_middleware"
    HOOK_CLASS = "AioHTTPMiddleware"
    HOOK_METHOD = "handle"

    def __init__(
        self,
        strategy_key,
        observation_queue,
        queue,
        import_hook,
        before_hook_point=None,
    ):
        from ..middlewares.aiohttp_middleware import AioHTTPMiddleware

        super(AioHTTPHookStrategy, self).__init__(
            strategy_key,
            observation_queue,
            queue,
            import_hook,
            before_hook_point,
        )
        self.middleware = AioHTTPMiddleware(self, observation_queue, queue)
        self.wrapper = self._apply_middleware

    @staticmethod
    def _apply_middleware(original, middleware):
        return middleware(original)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Generic WSGI HTTP Request / Response stuff
"""
import json
import logging
import sys

from ..utils import Iterable, Mapping, to_unicode_safe
from .base import BaseRequest, BaseResponse

if sys.version_info[0] >= 3:
    from http.cookies import SimpleCookie
    from urllib.parse import parse_qs, quote
else:
    from Cookie import SimpleCookie
    from urlparse import parse_qs
    from urllib import quote


LOGGER = logging.getLogger(__name__)


class BaseWSGIRequest(BaseRequest):
    """ Base class for WSGI based framework requests.
    """

    @property
    def form_params(self):
        # TODO: use a fallback form parser (like FormDataParser from werkzeug)
        # to parse the body when form params cannot be retrieved from the
        # framework.
        # Returns an empty dict for now.
        return {}

    @property
    def query_params(self):
        raw_query = self.get_raw_header("QUERY_STRING", "")
        try:
            return parse_qs(raw_query)
        except Exception:
            LOGGER.debug(
                "couldn't parse URL query %s", raw_query, exc_info=True
            )
            return {}

    @property
    def view_params(self):
        return {}

    @property
    def json_params(self):
        try:
            ct = to_unicode_safe(self.content_type)
            data = self.body
            if ct is not None and ct.startswith("application/json") and data:
                try:
                    return json.loads(data)
                except TypeError:
                    # Python 3.5 does not detect argument encoding, enforce UTF-8
                    return json.loads(to_unicode_safe(data))
        except Exception:
            LOGGER.debug("couldn't parse input json", exc_info=True)
        return {}

    @property
    def cookies_params(self):
        cookies = self.get_raw_header("HTTP_COOKIES")
        if cookies:
            try:
                cookie = SimpleCookie()
                cookie.load(cookies)
                return {key: cookie[key].coded_value for key in cookie.keys()}
            except Exception:
                LOGGER.debug("couldn't parse cookies", exc_info=True)
        return {}

    @property
    def body(self):
        from ..runtime_storage import runtime

        storage = self.storage or runtime
        body, _ = storage.get_request_input_preview()
        return body

    @property
    def remote_addr(self):
        """Remote IP address."""
        return to_unicode_safe(self.get_raw_header("REMOTE_ADDR"))

    @property
    def hostname(self):
        return to_unicode_safe(self.get_raw_header("HTTP_HOST", self.get_raw_header("SERVER_NAME")))

    @property
    def path(self):
        return quote(self.get_raw_header("SCRIPT_NAME", "")) + quote(
            self.get_raw_header("PATH_INFO", "")
        )

    @property
    def method(self):
        return to_unicode_safe(self.get_raw_header("REQUEST_METHOD"))

    @property
    def client_user_agent(self):
        return to_unicode_safe(self.get_raw_header("HTTP_USER_AGENT"))

    @property
    def referer(self):
        return to_unicode_safe(self.get_raw_header("HTTP_REFERER"))

    @property
    def scheme(self):
        return to_unicode_safe(self.get_raw_header("wsgi.url_scheme"))

    @property
    def server_port(self):
        return to_unicode_safe(self.get_raw_header("SERVER_PORT"))

    @property
    def remote_port(self):
        return to_unicode_safe(self.get_raw_header("REMOTE_PORT"))


class WSGIRequest(BaseWSGIRequest):
    """ Helper around raw wsgi environ
    """

    def __init__(self, environ, storage=None):
        super(WSGIRequest, self).__init__(storage=storage)
        self.environ = environ

    @property
    def raw_headers(self):
        return self.environ


class WSGIResponse(BaseResponse):
    """Helper around raw WSGI response."""

    def __init__(self, status, response_headers={}, body=None):
        self.status = status
        self.body = body
        if isinstance(response_headers, Mapping):
            self.headers = response_headers
        elif isinstance(response_headers, Iterable):
            self.headers = dict(response_headers)
        else:
            raise ValueError("Unknown WSGI response header type")

    @property
    def status_code(self):
        try:
            status = self.status
            if isinstance(self.status, bytes):
                status = status.decode("latin_1")
            return int(status.split(" ", 1)[0])
        except Exception:
            LOGGER.debug("cannot parse HTTP status", exc_info=True)
            return None

    @property
    def content_type(self):
        return self.headers.get("Content-Type")

    @property
    def content_length(self):
        try:
            return int(self.headers.get("Content-Length"))
        except (ValueError, TypeError):
            return None



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Instrumentation helpers
"""
import threading

LOCAL = threading.local()


def guard_call(key, callback, *args, **kwargs):
    """ Conditionnaly call callback checking that it cannot be called
    recursively
    """
    # Set guard_call_set if not present, required for multi-threaded env
    if not hasattr(LOCAL, "guard_call_set"):
        LOCAL.guard_call_set = set()

    if key in LOCAL.guard_call_set:
        return {"status": "not_executed"}

    LOCAL.guard_call_set.add(key)

    try:
        output = callback(*args, **kwargs)
        return output
    finally:
        try:
            LOCAL.guard_call_set.remove(key)
        except KeyError:
            pass



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
import codecs
import datetime
import json
import sys

if sys.version_info >= (3, 5):
    from typing import Mapping, Iterable

    string_type = str
elif sys.version_info[0] >= 3:
    from collections.abc import Mapping, Iterable

    string_type = str
else:
    from collections import Mapping, Iterable

    string_type = basestring  # noqa


def codecs_error_ascii_to_hex(exception):
    """On unicode decode error (bytes -> unicode error), tries to replace
    invalid unknown bytes by their hex notation."""
    if isinstance(exception, UnicodeDecodeError):
        obj = exception.object
        start = exception.start
        end = exception.end

        invalid_part = obj[start:end]
        result = []

        for character in invalid_part:
            # Python 2 strings
            if isinstance(character, str):
                result.append(u"\\x{}".format(character.encode("hex")))
            # Python 3 int
            elif isinstance(character, int):
                result.append(u"\\{}".format(hex(character)[1:]))
            else:
                raise exception
        result = ("".join(result), end)
        return result
    raise exception


codecs.register_error("__sqreen_ascii_to_hex", codecs_error_ascii_to_hex)


class CustomJSONEncoder(json.JSONEncoder):
    """Custom JsonEncoder which can handle datetime objects."""

    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        elif isinstance(obj, bytes):
            return obj.decode("utf-8", errors="__sqreen_ascii_to_hex")
        else:
            try:
                return repr(obj)
            except Exception:
                return "instance of type {}".format(repr(obj.__class__))


def reencode_payload(payload):
    """Do everything necessary to be able to encode a payload into JSON."""
    if isinstance(payload, (bytes, string_type)):
        return _reencode_string(payload)
    elif isinstance(payload, Mapping):
        return {
            _reencode_string(key): reencode_payload(value)
            for key, value in payload.items()
        }
    elif isinstance(payload, Iterable):
        return [reencode_payload(item) for item in payload]
    return payload


def _reencode_string(string):
    """Ensure that the string is encodable into JSON."""
    if isinstance(string, bytes):
        return string.decode("utf-8", errors="__sqreen_ascii_to_hex")
    return string



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Custom strategy for psycopg2 oddities
"""
import logging

from .dbapi2 import CustomConnection, CustomCursor
from .import_hook import ImportHookStrategy

LOGGER = logging.getLogger(__name__)


def wrap_custom_register_type(original):
    def custom_register_type(obj, scope=None):
        if isinstance(scope, (CustomConnection, CustomCursor)):
            scope = scope.__wrapped__
        return original(obj, scope)

    return custom_register_type


class Psycopg2Strategy(ImportHookStrategy):
    """ Simple strategy that replace psycopg2.*.register_type to works with our
    Proxy objects.
    """

    def import_hook_callback(self, original):
        """ Monkey-patch the object located at psycopg2.*.register_type to
        monkey-patch the register_type function
        """
        return wrap_custom_register_type(original)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" WSGI hook strategy
"""
from functools import partial

from ..._vendors.wrapt import FunctionWrapper
from ..middlewares.wsgi_middleware import WSGIMiddleware
from .import_hook import ImportHookStrategy


class WSGIStrategy(ImportHookStrategy):
    """
    Strategy to wrap a WSGI application.
    """

    def import_hook_callback(self, original):
        middleware = WSGIMiddleware(self, self.observation_queue, self.queue)

        def wrapper(wrapped, instance, args, kwargs):
            return middleware.wrap_app(wrapped, *args)

        return FunctionWrapper(original, wrapper)


class WSGIReceiverStrategy(ImportHookStrategy):
    """
    Strategy to wrap a WSGI application passed as the first argument of a
    function call.
    """

    def import_hook_callback(self, original):
        middleware = WSGIMiddleware(self, self.observation_queue, self.queue)

        def wrapper(wrapped, instance, args, kwargs):
            if args and callable(args[0]):
                new_args = list(args)
                new_args[0] = partial(middleware.wrap_app, args[0])
                args = new_args
            return wrapped(*args, **kwargs)

        return FunctionWrapper(original, wrapper)


class WSGIFactoryStrategy(ImportHookStrategy):
    """
    Strategy to wrap a WSGI application factory.
    """

    def import_hook_callback(self, original):
        middleware = WSGIMiddleware(self, self.observation_queue, self.queue)

        def wrapper(wrapped, instance, args, kwargs):
            app = wrapped(*args, **kwargs)
            return partial(middleware.wrap_app, app)

        return FunctionWrapper(original, wrapper)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Look for 404 error
"""
import logging
import os

from ..exceptions import InvalidArgument
from ..rules import RuleCallback
from ..utils import Mapping

LOGGER = logging.getLogger(__name__)


class NotFoundCB(RuleCallback):
    """Record attack for suspicious not found errors."""

    def __init__(self, *args, **kwargs):
        super(NotFoundCB, self).__init__(*args, **kwargs)

        if not isinstance(self.data, Mapping):
            msg = "Invalid data type received: {}"
            raise InvalidArgument(msg.format(type(self.data)))

        try:
            self.ignore_extensions = {'.{}'.format(e) for e in self.data['values'][0]["extensions"]}
        except KeyError as missing:
            msg = "mising key in data (had {}) {!r}"
            raise InvalidArgument(msg.format(self.data.keys(), missing))

        self.max_extensions_size = max([len(e) for e in self.ignore_extensions])

    def post(self, instance, args, kwargs, **options):
        request = self.storage.get_current_request()
        response = self.storage.get_current_response()
        if response is not None and response is not None \
                and response.status_code == 404:
            path = request.path
            _, maybe_ext = os.path.splitext(path)
            if len(maybe_ext) > self.max_extensions_size \
                    or maybe_ext.lower() not in self.ignore_extensions:
                infos = {
                    "path": path,
                    "host": request.hostname,
                    "verb": request.method,
                    "ua": request.client_user_agent,
                }
                self.record_attack(infos)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Django specific WSGI HTTP Request / Response stuff
"""
from logging import getLogger

import pkg_resources

from ..utils import to_unicode_safe
from .base import BaseResponse
from .wsgi import BaseWSGIRequest

LOGGER = getLogger(__name__)


def drf_preload_data():
    """Detect if request data should be preloaded for Django Rest Framework.

    See https://github.com/encode/django-rest-framework/issues/3951.
    """
    for package in pkg_resources.working_set:
        if package.project_name == "djangorestframework":
            drf_version_info = tuple(map(int, package.version.split(".")[:2]))
            return (3, 3) <= drf_version_info <= (3, 4)
    return False


class DjangoRequest(BaseWSGIRequest):

    DRF_PRELOAD_DATA = drf_preload_data()

    def __init__(self, request, storage=None):
        super(DjangoRequest, self).__init__(storage=storage)
        self.request = request

        # Cache for params
        self._query_params = None
        self._form_params = None
        self._cookies_params = None

    @property
    def query_params(self):
        if self._query_params is None:
            try:
                # Convert django QueryDict to a normal dict with values as list
                self._query_params = dict(self.request.GET.lists())
            except Exception:
                LOGGER.debug("couldn't get request.GET from the framework",
                             exc_info=True)
                self._query_params = super(DjangoRequest, self).query_params
        return self._query_params

    @property
    def body(self):
        try:
            return self.request.body
        except Exception:  # UnreadablePostError, RawPostDataException, RequestDataTooBig
            LOGGER.debug("couldn't get request.body from the framework",
                         exc_info=True)
            return super(DjangoRequest, self).body

    @property
    def form_params(self):
        if self._form_params is None:
            try:
                if (
                    self.DRF_PRELOAD_DATA
                    and self.request.method == "POST"
                    and not getattr(self.request, "_read_started", False)
                ):
                    self.request.body
                self._form_params = dict(self.request.POST)
            except Exception:
                # Django can raise ValueError or TypeError exceptions when
                # parsing the HTTP body for reading POST data.
                LOGGER.debug("couldn't get request.POST from framework",
                             exc_info=True)
                self._form_params = super(DjangoRequest, self).form_params
        return self._form_params

    @property
    def cookies_params(self):
        if self._cookies_params is None:
            try:
                self._cookies_params = self.request.COOKIES
            except Exception:
                LOGGER.debug("couldn't get request.COOKIES from framework",
                             exc_info=True)
                self._cookies_params = super(DjangoRequest, self).cookies_params
        return self._cookies_params

    @property
    def remote_addr(self):
        return to_unicode_safe(self.get_raw_header("REMOTE_ADDR"))

    @property
    def hostname(self):
        try:
            return self.request.get_host()
        except Exception:
            LOGGER.debug("couldn't get request.get_host from the framework",
                         exc_info=True)
            return None

    @property
    def method(self):
        return self.request.method

    @property
    def referer(self):
        return to_unicode_safe(self.get_raw_header("HTTP_REFERER"))

    @property
    def client_user_agent(self):
        return to_unicode_safe(self.get_raw_header("HTTP_USER_AGENT"))

    @property
    def route(self):
        """Request route."""
        # The resolver_match attribute is set only set once all request
        # middlewares have been called.
        resolver_match = getattr(self.request, "resolver_match", None)
        return getattr(resolver_match, "view_name", None)

    @property
    def path(self):
        return self.request.path

    @property
    def scheme(self):
        return getattr(self.request, "scheme", None)

    @property
    def server_port(self):
        return to_unicode_safe(self.get_raw_header("SERVER_PORT"))

    @property
    def remote_port(self):
        return to_unicode_safe(self.get_raw_header("REMOTE_PORT"))

    @property
    def raw_headers(self):
        return self.request.META

    @property
    def view_params(self):
        resolver_match = getattr(self.request, "resolver_match", None)
        if resolver_match:
            return resolver_match.kwargs


class DjangoResponse(BaseResponse):

    def __init__(self, response):
        self.response = response

    @property
    def status_code(self):
        return self.response.status_code

    @property
    def content_type(self):
        return self.response.get("Content-Type")

    @property
    def content_length(self):
        try:
            return int(self.response.get("Content-Length"))
        except (ValueError, TypeError):
            return None



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
import logging
import traceback
from datetime import datetime

from ..actions import BaseAction
from ..runtime_storage import RuntimeStorage, runtime
from ..utils import HAS_TYPING, is_string, naive_dt_to_utc, now

if HAS_TYPING:
    from typing import Mapping, Optional, Set


LOGGER = logging.getLogger(__name__)


_SQREEN_EVENT_PREFIX = "sq."

_TRACK_OPTIONS_FIELDS = frozenset(
    ["properties", "user_identifiers", "timestamp"]
)

_TRACK_PAYLOAD_SECTIONS = ("request", "response", "params", "headers")

STACKTRACE_EVENTS = set()  # type: Set[str]

_MAX_EVENT_PROPERTIES = 16


def _track_unsafe(event, options, storage=runtime):
    # type: (str, Mapping, RuntimeStorage) -> bool
    """Track an SDK event.

    This function is used internally in the agent to send built-in SDK events,
    e.g. output of security actions. It does not perform any check and is not
    exposed to the user.
    """
    if "timestamp" not in options:
        options = dict(options)
        options["timestamp"] = now()
    storage.observe(
        "sdk",
        ["track", options["timestamp"], event, options],
        payload_sections=_TRACK_PAYLOAD_SECTIONS,
        report=True,
    )
    return True


def track(event, options=None, storage=runtime):
    # type: (str, Optional[Mapping], RuntimeStorage) -> bool
    """Track an SDK event."""
    # Check event type.
    if not is_string(event):
        raise TypeError(
            "event name must be a string, not {}".format(
                event.__class__.__name__
            )
        )
    # Check event name.
    if event.startswith(_SQREEN_EVENT_PREFIX):
        LOGGER.warning(
            "Event names starting with %r are reserved, "
            "event %r has been ignored",
            _SQREEN_EVENT_PREFIX,
            event,
        )
        return False
    if options is None:
        options = {}
    else:
        options = dict(options)
    # Check option keys.
    for option_key in list(options):
        if option_key not in _TRACK_OPTIONS_FIELDS:
            LOGGER.warning("Invalid option key %r, skipped", option_key)
            del options[option_key]
    timestamp = options.get("timestamp")
    if timestamp is not None:
        if not isinstance(timestamp, datetime):
            raise TypeError(
                "timestamp option must be a datetime object, not {}".format(
                    event.__class__.__name__
                )
            )
        if timestamp.tzinfo is None:
            LOGGER.info("Event %r timestamp is not timezone-aware, default to UTC",
                        event)
            options["timestamp"] = timestamp = naive_dt_to_utc(timestamp)

    properties = options.get("properties")
    # Check the number of properties.
    if properties and len(properties) > _MAX_EVENT_PROPERTIES:
        LOGGER.warning(
            "Event %r has %d properties, "
            "only the first %d ones will be reported",
            event,
            len(properties),
            _MAX_EVENT_PROPERTIES,
        )
        options["properties"] = dict(
            sorted(properties.items())[:_MAX_EVENT_PROPERTIES]
        )
    # Store stacktrace if required.
    if event in STACKTRACE_EVENTS:
        LOGGER.debug("Stacktrace recorded by for event %s", event)
        options["stacktrace"] = traceback.format_stack()
    return _track_unsafe(event, options, storage=storage)


def track_action(action, output, storage=runtime):
    # type: (BaseAction, Mapping, RuntimeStorage) -> bool
    """Track an action output."""
    if not action.send_response:
        return False
    return _track_unsafe(
        "sq.action.{}".format(action.name),
        {"properties": {"output": output, "action_id": action.iden}},
        storage=storage,
    )



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Record a WSGI request and enforce the whitelist.
"""

from logging import getLogger

from ..frameworks.wsgi import WSGIRequest
from ..rules import RuleCallback

LOGGER = getLogger(__name__)


class RecordRequest(RuleCallback):

    INTERRUPTIBLE = False

    @property
    def whitelisted(self):
        return False

    def pre(self, instance, args, kwargs, **options):
        self._store_request(WSGIRequest(args[0]))

    def _store_request(self, request):
        current_request = self.storage.get_current_request()
        trace_request = self.performance_monitoring_enabled \
            and current_request is None
        # If a request is already stored, we probably want to upgrade the
        # request object from a WSGIRequest to a DjangoRequest or another
        # framework. The request trace was already started, there is no need to
        # start a new one.
        self.storage.store_request(request, trace_request=trace_request)

        runner = self.runner
        if runner is not None and hasattr(request, "is_debug") \
                and runner.settings.get_debug_flag() is None:
            # Set the debug flag on the runner if possible and not already set.
            runner.settings.set_debug_flag(request.is_debug())

        whitelist_match = self.storage.get_whitelist_match()
        if whitelist_match is not None:
            return
        if runner is not None:
            whitelist_match = runner.settings.whitelist_match(request)
            if whitelist_match is not None and runner.whitelisted_metric:
                self.record_observation("whitelisted", whitelist_match, 1)
        # Set to False if the request is not whitelisted, in order to avoid
        # further useless lookups.
        self.storage.set_whitelist_match(whitelist_match or False)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Insert X-Protected-By header
"""
import logging

from ..exceptions import InvalidArgument
from ..rules import RuleCallback
from ..utils import Mapping

LOGGER = logging.getLogger(__name__)


def convert_to_str(headers):
    """ Encode a list of headers tuples into latin1
    """
    for header_name, header_value in headers:
        header_name = str(
            header_name.encode("latin-1", errors="replace").decode("latin-1")
        )
        header_value = str(
            header_value.encode("latin-1", errors="replace").decode("latin-1")
        )
        yield (header_name, header_value)


class BaseHeadersInsertCB(RuleCallback):
    """ Base class for header insertion callbacks
    """

    def __init__(self, *args, **kwargs):
        super(BaseHeadersInsertCB, self).__init__(*args, **kwargs)

        if not isinstance(self.data, Mapping):
            msg = "Invalid data type received: {}"
            raise InvalidArgument(msg.format(type(self.data)))

        try:
            self.values = self.data["values"]
        except KeyError:
            msg = "No key 'values' in data (had {})"
            raise InvalidArgument(msg.format(self.data.keys()))

        self._headers = None

    @property
    def headers(self):
        """ Cached property to defer headers data conversio,n
        """
        if self._headers is None:
            self._headers = dict(convert_to_str(self.values))
        return self._headers


class HeadersInsertCB(BaseHeadersInsertCB):
    """ Callback that add the custom sqreen header in WSGI
    """

    def pre(self, instance, args, kwargs, **options):
        new_args = list(args)
        headers_to_add = dict(self.headers)
        new_headers = []
        current_headers = args[1]
        for item in current_headers:
            header = headers_to_add.pop(item[0], None)
            if header is not None:
                new_headers.append((item[0], header))
            else:
                new_headers.append(item)
        for item in headers_to_add.items():
            new_headers.append(item)

        new_args[1] = new_headers
        return {"status": "modify_args", "args": [new_args, kwargs]}



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Thread-safe samplers
"""

import itertools
import random
import threading
from collections import namedtuple
from datetime import timedelta

from .utils import HAS_TYPING, now

if HAS_TYPING:
    from datetime import datetime
    from typing import Any, Dict, Optional, Sequence


class Decision(namedtuple("Decision", ["sampled", "attrs"])):

    __slots__ = ()

    def __new__(cls, sampled, attrs=None):  # type: (bool, Optional[Dict]) -> "Decision"
        if attrs is None:
            attrs = {}
        return super(Decision, cls).__new__(cls, sampled, attrs)

    def __nonzero__(self):
        return self.sampled

    def __bool__(self):
        return self.sampled


class BaseSampler(object):

    def should_sample(self, **kwargs):  # type (**Any) -> Decision
        raise NotImplementedError


class CallSampler(BaseSampler):

    def __init__(self, n_calls):  # type (int) -> None
        self.n_calls = n_calls
        self._state = itertools.count()

    def should_sample(self, **kwargs):  # type (**Any) -> Decision
        state = next(self._state)
        return Decision(state % self.n_calls == 0, dict(counter=state))


class MaxCallSampler(BaseSampler):

    def __init__(self, max_calls):  # type (int) -> None
        self.max_calls = max_calls
        self._state = itertools.count()

    def should_sample(self, **kwargs):  # type (**Any) -> Decision
        state = next(self._state)
        return Decision(state < self.max_calls, dict(counter=state))


class ProbabilitySampler(BaseSampler):

    def __init__(self, rate):  # type: (float) -> None
        if rate < 0:
            raise ValueError("rate cannot be negative")
        if rate > 1:
            rate = 1.0
        self.rate = rate

    def should_sample(self, random_value=None, **kwargs):  # type (Optional[float], **Any) -> Decision
        if random_value is None:
            random_value = random.random()
        return Decision(random_value < self.rate, dict(random_value=random_value))


class MaxCallWindowSampler(MaxCallSampler):

    def __init__(self, max_calls, window_duration_s):  # type: (int, float) -> None
        super(MaxCallWindowSampler, self).__init__(max_calls)
        self.window_duration_s = timedelta(seconds=window_duration_s)
        self._window_started_at = None
        self._lock = threading.Lock()

    def should_sample(self, at=None, **kwargs):  # type (Optional[datetime], **Any) -> Decision
        with self._lock:
            if at is None:
                at = now()
            if self._window_started_at is None or at - self._window_started_at >= self.window_duration_s:
                self._window_started_at = at
                self._state = itertools.count()
            attrs = dict(window_started_at=self._window_started_at, at=at)
        decision = super(MaxCallWindowSampler, self).should_sample(**kwargs)
        decision.attrs.update(attrs)
        return decision


class MaxDurationSampler(BaseSampler):

    def __init__(self, max_duration_s, at=None):  # type: (float, Optional[datetime]) -> None
        if at is None:
            at = now()
        self._max_duration_at = at + timedelta(seconds=max_duration_s)

    def should_sample(self, at=None, **kwargs):  # type: (Optional[datetime], **Any) -> Decision
        if at is None:
            at = now()
        return Decision(at < self._max_duration_at, dict(at=at))


class AnySampler(BaseSampler):

    def __init__(self, samplers):  # type: (Sequence[BaseSampler]) -> None
        self.samplers = samplers

    def should_sample(self, **kwargs):  # type: (**Any) -> Decision
        try:
            idx, decision = next(
                (idx, decision) for idx, decision in
                enumerate(sampler.should_sample(**kwargs) for sampler in self.samplers)
                if decision
            )
            return Decision(True, dict(idx=idx, child=decision))
        except StopIteration:
            return Decision(False)


class AllSampler(BaseSampler):

    def __init__(self, samplers):  # type: (Sequence[BaseSampler]) -> None
        self.samplers = samplers

    def should_sample(self, **kwargs):  # type: (**Any) -> Decision
        children = list([sampler.should_sample(**kwargs) for sampler in self.samplers])
        return Decision(all(children), dict(children=children))


class LineTracingSampler(AllSampler):

    def __init__(self, line):
        samplers = []
        line = dict(line)
        n_calls = line.pop("calls", None)
        if n_calls is not None:
            samplers.append(CallSampler(n_calls))
        rate = line.pop("random", None)
        if rate is not None:
            samplers.append(ProbabilitySampler(rate))
        max_duration_m = line.pop("max_duration_minutes", None)
        if max_duration_m is not None:
            samplers.append(MaxDurationSampler(max_duration_m * 60))
        target_per_m = line.pop("target_per_minute", None)
        if target_per_m is not None:
            samplers.append(MaxCallWindowSampler(target_per_m, 60))
        max_calls = line.pop("max_calls", None)
        if max_calls is not None:
            samplers.append(MaxCallSampler(max_calls))
        if line:
            # Unknown primitives remaining, always False
            samplers.insert(0, AnySampler({}))
        super(LineTracingSampler, self).__init__(samplers)


class ScopedTracingSampler(AnySampler):

    def __init__(self, definition):
        enabled = definition.get("enabled")
        if enabled:
            self.instruction = definition.get("sampling", [{}])
            samplers = [LineTracingSampler(line) for line in self.instruction]
        else:
            self.instruction = []
            samplers = []
        super(ScopedTracingSampler, self).__init__(samplers)

    def should_sample(self, **kwargs):
        decision = super(ScopedTracingSampler, self).should_sample(**kwargs)
        if decision:
            decision.attrs["trigger"] = self.instruction[decision.attrs["idx"]]
        return decision


class TracingSampler(BaseSampler):

    def __init__(self, scope_definition={}):
        scope_definition = dict(scope_definition)
        default_definition = scope_definition.pop("*", {})
        self.default_sampler = ScopedTracingSampler(default_definition)
        self.samplers = {scope: ScopedTracingSampler(definition)
                         for scope, definition in scope_definition.items()}

    def should_sample(self, scope=None, **kwargs):  # type: (Optional[str], **Any) -> Decision
        sampler = self.samplers.get(scope)
        if sampler is not None:
            return sampler.should_sample(**kwargs)
        return self.default_sampler.should_sample(**kwargs)

    def update_scope_definition(self, scope_definition):
        """ Update the current scope definition with a new scope definition.
        """
        obj = self.__class__(scope_definition)
        new_samplers = dict(self.samplers)
        new_samplers.update(obj.samplers)
        if "*" in scope_definition:
            self.default_sampler = obj.default_sampler
        self.samplers = new_samplers



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Count http status codes
"""
from ..rules import RuleCallback


class CountHTTPCodesCB(RuleCallback):

    INTERRUPTIBLE = False

    def post(self, instance, args, kwargs, **options):
        response = self.storage.get_current_response()
        if response is not None:
            self.record_observation("http_code", str(response.status_code), 1)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Pyramid hook strategy
"""

from logging import getLogger

from ..middlewares.pyramid_middleware import PyramidMiddleware
from .framework import FrameworkStrategy

LOGGER = getLogger(__name__)


def tween_call_insert(original, middleware):
    def wrapped_tween_call(self, handler, *args, **kwargs):
        LOGGER.debug("Execute tween_call_insert")
        new_handler = middleware(handler)

        return original(self, new_handler, *args, **kwargs)

    return wrapped_tween_call


class PyramidStrategy(FrameworkStrategy):
    """ Strategy for Pyramid peripheric callbacks.

    It injects a custom PyramidMiddleware that calls callbacks for each
    lifecycle method
    """

    MODULE_NAME = "pyramid.config.tweens"
    HOOK_CLASS = "Tweens"
    HOOK_METHOD = "__call__"

    def __init__(
        self,
        strategy_key,
        observation_queue,
        queue,
        import_hook,
        before_hook_point=None,
    ):
        super(PyramidStrategy, self).__init__(
            strategy_key,
            observation_queue,
            queue,
            import_hook,
            before_hook_point,
        )

        self.middleware = PyramidMiddleware(self, observation_queue, queue)
        self.wrapper = tween_call_insert



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
import sys

from ..._vendors.wrapt import FunctionWrapper
from .base import BaseMiddleware


class FlaskMiddleware(BaseMiddleware):
    def __call__(self, original):

        def wrapper(wrapped, instance, args, kwargs):
            """ Call the lifecycles methods with these arguments:
            Flask pre callbacks will receive these arguments:
            (None)
            Flask post callbacks will receive these arguments:
            (None, response)
            Flask failing callbacks will receive these arguments:
            (None, exception)
            """
            self.strategy.before_hook_point()

            self.execute_pre_callbacks(record_attack=True)

            try:
                response = wrapped(*args, **kwargs)
            except Exception:
                self.execute_failing_callbacks(sys.exc_info())
                raise
            return self.execute_post_callbacks(
                response, record_attack=True
            )

        return FunctionWrapper(original, wrapper)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Custom error page for Flask
"""
from .headers_insert import convert_to_str
from .sqreen_error_page import BaseSqreenErrorPage


class SqreenErrorPageFlask(BaseSqreenErrorPage):

    def pre(self, instance, args, kwargs, **options):

        exception = args[0]
        ret = self.handle_exception(exception)

        if ret is not None:
            status_code, content, headers = ret

            from flask import make_response

            response = make_response(content)
            response.status_code = status_code

            for header_name, header_value in convert_to_str(headers.items()):
                response.headers[header_name] = header_value

            return {
                "status": "override",
                "new_return_value": response
            }



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Abstract callback for checking regex against request fields
"""

from logging import getLogger

from ..binding_accessor import BindingAccessor
from ..exceptions import InvalidArgument
from ..matcher import Matcher
from ..performance_metrics import Trace
from ..rules import RuleCallback
from ..utils import Mapping, is_string

LOGGER = getLogger(__name__)


class BindingAccessorMatcherCallback(RuleCallback):

    SUPPORT_BUDGETS = True
    MAX_LEN = 1024 * 128

    def __init__(self, *args, **kwargs):
        super(BindingAccessorMatcherCallback, self).__init__(*args, **kwargs)

        if not isinstance(self.data, Mapping):
            msg = "Invalid data type received: {}"
            raise InvalidArgument(msg.format(type(self.data)))

        try:
            self.values = self.data["values"]
        except KeyError:
            msg = "No key 'values' in data (had {})"
            raise InvalidArgument(msg.format(self.data.keys()))

        self.patterns = None

    def _prepare_patterns(self):
        """ Prepare patterns if not ready
        """
        patterns = []

        ba_cache = {}

        def cached_binding_accessor(expression):
            if expression in ba_cache:
                return ba_cache[expression]
            else:
                binding_accessor = BindingAccessor(expression)
                ba_cache[expression] = binding_accessor
                return binding_accessor

        try:
            for value in self.values:

                bindings = [
                    cached_binding_accessor(ba)
                    for ba in value["binding_accessor"]
                ]

                patterns.append(
                    {
                        "id": value["id"],
                        "binding_accessor": bindings,
                        "matcher": Matcher([value["matcher"]]),
                    }
                )
        finally:
            self.patterns = patterns

    def pre(self, instance, args, kwargs, **options):
        request = self.storage.get_current_request()
        response = self.storage.get_current_response()

        if not request:
            LOGGER.warning("No request was recorded abort")
            return

        # Only compile patterns if we need them
        if self.patterns is None:
            self._prepare_patterns()

        budget = self.get_remaining_budget(options)

        binding_eval_args = {
            "request": request,
            "response": response,
            "inst": instance,
            "args": self.storage.get_current_args(args),
            "kwargs": kwargs,
            "data": self.data,
        }

        cache = {}
        total_pattern_time = 0
        for pattern in self.patterns:
            with Trace() as pattern_trace:
                for binding_accessor in pattern["binding_accessor"]:
                    expression = binding_accessor.expression

                    if expression in cache:
                        data = cache[expression]
                    else:
                        data = binding_accessor.resolve(**binding_eval_args)

                        # Convert string to list
                        if data is None or is_string(data):
                            data = [data]

                        # Ignore not string values as we use a string matcher
                        # that only match strings
                        data = [elem for elem in data if is_string(elem)]

                        cache[expression] = data

                    if not data:
                        continue

                    for elem in data:
                        if len(elem) > self.MAX_LEN:
                            continue

                        if pattern["matcher"].match(elem):
                            infos = {
                                "id": pattern["id"],
                                "binding_accessor": expression,
                                "matcher": pattern["matcher"].patterns,
                                "found": elem,
                            }
                            self.record_attack(infos)

                            # Potentially raise an attack
                            return {
                                "status": "raise",
                                "data": elem,
                                "rule_name": self.rule_name,
                            }
            total_pattern_time += pattern_trace.duration_ms
            if budget is not None and total_pattern_time >= budget:
                break



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Binding accessor class."""

import collections
import re
import sys
import types

from ._vendors.urllib3.packages.six import reraise  # type: ignore
from .utils import HAS_TYPING, flatten

if HAS_TYPING:
    from typing import (
        Any,
        Callable,
        Dict,
        Iterable,
        List,
        Mapping,
        Match,
        Optional,
        Pattern,
        Sequence,
        Tuple,
        Union,
    )

    ExcType = Tuple[type, BaseException, types.TracebackType]


def transform_len(value, exc=None, params=[], **kwargs):
    # type: (Any, Optional[ExcType], Sequence[BindingAccessorPart], **Any) -> Any
    """Return the length of the iterable."""
    if exc:
        reraise(*exc)
    return len(value)


def flat_keys(value, exc=None, params=[], **kwargs):
    # type: (Any, Optional[ExcType], Sequence[BindingAccessorPart], **Any) -> Any
    """Return the length of the iterable."""
    """Return the list of keys in iterable and nested iterables."""
    if exc:
        reraise(*exc)
    max_iterations = params[0].resolve(**kwargs) if params else 1000
    keys, _ = flatten(value, max_iterations=max_iterations)
    return keys


def flat_values(value, exc=None, params=[], **kwargs):
    # type: (Any, Optional[ExcType], Sequence[BindingAccessorPart], **Any) -> Any
    """Return the list of values in iterable and nested iterables."""
    if exc:
        reraise(*exc)
    max_iterations = params[0].resolve(**kwargs) if params else 1000
    _, values = flatten(value, max_iterations=max_iterations)
    return values


def transform_default(value, exc=None, params=[], **kwargs):
    # type: (Any, Optional[ExcType], Sequence[BindingAccessorPart], **Any) -> Any
    """Resolve the parameter if the previous binding accessor resolution
    or transformation raised an exception else return the value as it is.
    """
    if exc:
        if params:
            return params[0].resolve(**kwargs)
        return None
    return value


def transform_type(value, exc=None, params=[], **kwargs):
    # type: (Any, Optional[ExcType], Sequence[BindingAccessorPart], **Any) -> Any
    """Get the type of a binding accessor resolved value.
    """
    if exc:
        reraise(*exc)
    return type(value).__name__


def transform_encode(value, exc=None, params=[], **kwargs):
    # type: (Any, Optional[ExcType], Sequence[BindingAccessorPart], **Any) -> Any
    """Transform to encode a value to bytes.
    """
    if exc:
        reraise(*exc)
    if isinstance(value, bytes):
        return value
    encoding = "utf-8"
    errors = "replace"
    if params:
        encoding = params[0].resolve(**kwargs)
        if len(params) > 1:
            errors = params[1].resolve(**kwargs)
    if not isinstance(value, str):
        value = str(value)
    return value.encode(encoding=encoding, errors=errors)


class StringScanner(object):
    """Lexical scanning operations on a string.

    This is a poor man equivalent of Ruby's StringScanner class.
    """

    __slots__ = ("string", "_size", "_pos", "match")

    def __init__(self, string):  # type: (str) -> None
        self.string = string
        self._size = len(string)
        self._pos = 0
        self.match = None  # type: Optional[Match[str]]

    def tell(self):  # type: () -> int
        """Return the current position in the string."""
        return self._pos

    def rewind(self, delta):  # type: (int) -> None
        """Move the cursor back `delta` characters.
        """
        self._pos -= delta

    def scan_string(self, sub):  # type: (str) -> Optional[str]
        """Find the given substring from the current position.

        If the substring is present, the cursor position is updated and the
        substring is returned. Otherwise, this method returns None.

        This is semantically equivalent to calling scan_regex with an exact
        regexp, but faster.
        """
        size = len(sub)
        if self.string[self._pos : self._pos + size] == sub:
            self._pos += size
            return sub
        return None

    def scan_regex(self, regex):
        # type: (Pattern[str]) -> Optional[Match[str]]
        """Match the regex from the current position.

        If there is a match, the cursor position is updated and the match
        object is returned. Otherwise, this method returns None.
        """
        self.match = regex.match(self.string, self._pos)
        if self.match is not None:
            self._pos += len(self.match.group(0))
            return self.match
        return None

    def scan_until(self, delimiter=None):
        # type: (Optional[str]) -> Iterable[None]
        """Helper for common scan until a delimiter pattern.

        Yield until the delimiter or the end of the string is reached.
        The generator also stops when the scanner does not advance the scanner.
        """
        pos = self.tell()
        while not self.finished() and not (delimiter and self.scan_string(delimiter)):
            yield
            if pos == self.tell():
                raise ValueError("parsing error, scanner is stuck")

    def scan_until_regex(self, regex):
        # type: (Pattern[str]) -> Iterable[None]
        """Same helper than scan_until but using a regular expression.
        """
        pos = self.tell()
        while not self.finished() and not self.scan_regex(regex):
            yield
            if pos == self.tell():
                raise ValueError("parsing error, scanner is stuck")

    def finished(self):  # type: () -> bool
        """True if the cursor is at the end of the string, False otherwise."""
        return self._pos == self._size


BindingAccessorComponent = collections.namedtuple(
    "BindingAccessorComponent", ["kind", "value"])


class BindingAccessorPart(object):
    """Binding accessor part.

    The class constructor is given a binding accessor (e.g. "#.args[0]")
    and parses it.
    """

    __slots__ = ("path", "expression", "_scanner")

    def __init__(self, expression=None, scanner=None):
        # type: (Optional[str], Optional[StringScanner]) -> None
        self.path = []  # type: List[BindingAccessorComponent]
        self._scanner = scanner
        self.expression = expression
        self._parse(expression)

    _END_OF_EXPRESSION = re.compile(r"[ \|\),]")

    def _parse(self, expression):  # type: (Optional[str]) -> None
        """Parse a binding accessor part.

        It is internally converted to a series of components that are later
        resolved.
        """
        if self._scanner is None:
            if expression is None:
                raise ValueError(
                    "either expression or scanner must be passed as arguments")
            self._scanner = StringScanner(expression)

        start_pos = self._scanner.tell()

        # Check for scalar values first then variables.
        scalar = self._scan_scalar()
        if scalar:
            self.path.append(scalar)
        else:
            self._scan_push_variable()

        # Scan until something looks like a transform operator
        # or another transform parameter
        for _ in self._scanner.scan_until_regex(self._END_OF_EXPRESSION):
            self._scan_push_attribute()
            self._scan_push_indexes()

        # Do not consume the last character because it might be a transform
        # operator.
        if not self._scanner.finished():
            self._scanner.rewind(1)

        # Keep the expression and delete the scanner instance.
        self.expression = self._scanner.string[start_pos:self._scanner._pos]
        self._scanner = None

    _SQREEN_VARIABLE_REGEX = re.compile(r"#\.(\w+)")
    _PYTHON_IDENTIFIER_REGEX = re.compile(r"[a-zA-Z_](?:\w)*")

    def _scan_push_variable(self):  # type:  () -> None
        """Try to parse a variable and push it to the components list.

        Do nothing if the scanner cannot parse a variable or a Python
        identifier.
        """
        assert self._scanner is not None, "parsing is done"
        match = self._scanner.scan_regex(self._SQREEN_VARIABLE_REGEX)
        if match:
            variable_name = match.group(1)
            if variable_name == "cargs":
                # Hack for the PHP daemon using cargs instead of args.
                variable_name = "args"
            self.path.append(BindingAccessorComponent(
                kind="sqreen-variable",
                value=variable_name,
            ))
        else:
            match = self._scanner.scan_regex(self._PYTHON_IDENTIFIER_REGEX)
            if match:
                self.path.append(BindingAccessorComponent(
                    kind="variable",
                    value=match.group(),
                ))

    def _scan_push_attribute(self):  # type:  () -> None
        """Try to parse an attribute and push it to the components list.

        Do nothing of the scanner cannot parse a Python identifier.
        """
        assert self._scanner is not None, "parsing is done"
        while self._scanner.scan_string(".") is not None:
            match = self._scanner.scan_regex(self._PYTHON_IDENTIFIER_REGEX)
            if match:
                self.path.append(BindingAccessorComponent(
                    kind="attribute",
                    value=match.group(),
                ))

    _SCALAR_REGEX = re.compile(
        r"(?P<integer>-?\d+)|'(?P<string>(?:\\.|[^\\'])*)'|"
        "(?P<none>[Nn]one)|(?P<boolean>([Tt]rue)|([Ff]alse))"
    )
    _SCALAR_COERCE_FUNCS = {
        "integer": int,
        "none": lambda x: None,
        "boolean": lambda x: x == "True",
    }  # type: Dict[str, Callable[[Any], Any]]

    def _scan_scalar(self):  # type:  () -> Optional[BindingAccessorComponent]
        """Scan a scalar value and return the corresponding components.

        Return None if the scanner cannot parse an integer or a string.
        """
        assert self._scanner is not None, "parsing is done"
        match = self._scanner.scan_regex(self._SCALAR_REGEX)
        if match:
            last = match.lastgroup
            assert last is not None, "a group should have matched"
            value = match.group(last)
            coerce_func = self._SCALAR_COERCE_FUNCS.get(last)
            if coerce_func:
                value = coerce_func(value)
            return BindingAccessorComponent(kind="scalar", value=value)
        return None

    def _scan_push_indexes(self):  # type:  () -> None
        """Scan a sequence of indexes and push them to the components list."""
        assert self._scanner is not None, "parsing is done"
        while self._scanner.scan_string("[") is not None:
            scalar = self._scan_scalar()
            if scalar is None:
                raise ValueError("invalid index")
            if self._scanner.scan_string("]") is None:
                raise ValueError("unfinished index")
            self.path.append(scalar._replace(kind="index"))

    def resolve(self, **kwargs):  # type: (**Any) -> Any
        """Given a context, resolve the expression and return the value."""

        value = None
        for component in self.path:
            value = self._resolve_component(value, component, **kwargs)
        self._validate_value(value)
        return value

    def _resolve_component(self, value, component, **kwargs):
        # type: (Any, BindingAccessorComponent, Any) -> Any
        """Resolve a component.

        Simple expressions (static strings, integers, indexes and
        attributes) are directly resolved. Others are dispatched to specialized
        methods.
        """
        component_kind = component.kind
        if component_kind == "sqreen-variable":
            return self._resolve_sqreen_variable(component.value, **kwargs)
        elif component_kind == "index":
            return value[component.value]
        elif component_kind == "scalar":
            return component.value
        elif component_kind == "attribute":
            return getattr(value, component.value)
        elif component_kind == "variable":
            return self._resolve_variable(component.value, **kwargs)
        raise ValueError(
            "invalid component kind {!r}".format(component_kind)
        )

    _NO_VALUE = object()
    _SQREEN_VARIABLES = frozenset(["args", "kwargs", "request", "response", "rv", "data", "inst"])

    def _resolve_sqreen_variable(self, what, **kwargs):
        # type: (str, Any) -> Any
        """Resolve sqreen-variables (the ones starting with #.).

        Fall back on the request object if the value is not a special
        sqreen-variable. Return None if the request is None.
        """
        value = kwargs.get(what, self._NO_VALUE)
        if value is not self._NO_VALUE:
            return value
        elif what in self._SQREEN_VARIABLES:
            return None
        request = kwargs.get("request")
        if request is not None:
            return getattr(request, what)

    def _resolve_variable(self, variable_name, binding={}, global_binding={},
                          **kwargs):
        # type: (str, Mapping, Mapping, Any) -> Any
        """Resolve a general variable name.

        Search in local context first, then in general context.
        """
        value = binding.get(variable_name, self._NO_VALUE)
        if value is self._NO_VALUE:
            value = global_binding.get(variable_name, self._NO_VALUE)
            if value is self._NO_VALUE:
                raise NameError("name {!r} was not found in bindings"
                                .format(variable_name))
        return value

    def _validate_value(self, value):  # type: (Any) -> None
        """Raise ValueError if the value is a method or a function."""
        if isinstance(value, (types.FunctionType, types.MethodType)):
            raise ValueError("invalid return value {!r}".format(value))

    @property
    def pretty_expression(self):  # type: () -> str
        """Format the expression in a human readable way."""
        expression = []
        for item in self.path:
            if item.kind == "sqreen-variable":
                expression.append("#.{}".format(item.value))
            elif item.kind == "index":
                expression.append("[{}]".format(repr(item.value)))
            elif item.kind == "scalar":
                expression.append(repr(item.value))
            elif item.kind == "attribute":
                expression.append(".{}".format(item.value))
            elif item.kind == "variable":
                expression.append(item.value)
        return "".join(expression)

    def __repr__(self):  # type: () -> str
        return "{}({!r})".format(self.__class__.__name__, self.pretty_expression)


BindingAccessorTransform = collections.namedtuple("BindingAccessorTransform",
                                                  ["name", "func", "params"])


class BindingAccessor(object):
    """Binding accessors expression.

    The class constructor is given a binding accessor expression
    (e.g. "#.args[0] | len") and parses it.
    """

    __slots__ = ("parts", "expression", "_scanner")

    _TRANSFORM_FUNCS = {
        "len": transform_len,
        "flat_keys": flat_keys,
        "flat_values": flat_values,
        "default": transform_default,
        "type": transform_type,
        "encode": transform_encode,
    }  # type: Mapping[str, Callable]

    def __init__(self, expression=None, scanner=None):
        # type: (Optional[str], Optional[StringScanner]) -> None
        """Format the expression in a human readable way."""
        self.parts = []  # type: List[Union[BindingAccessorPart, BindingAccessorTransform]]
        self._scanner = scanner
        self.expression = expression
        self._parse(expression)

    _TRANSFORM_DELIMITER_REGEX = re.compile(r"\s*\|\s*")

    def _parse(self, expression):  # type: (Optional[str]) -> None
        if self._scanner is None:
            if expression is None:
                raise ValueError(
                    "either expression or scanner must be passed as arguments")
            self._scanner = StringScanner(expression)

        start_pos = self._scanner.tell()

        # We should always have a binding accessor as first element
        self._scan_push_binding_accessor()

        # Then, parse the transforms
        for _ in self._scanner.scan_until():
            if self._scanner.scan_regex(self._TRANSFORM_DELIMITER_REGEX):
                self._scan_push_transform()

        # Keep the expression and free the scanner
        self.expression = self._scanner.string[start_pos:self._scanner._pos]
        self._scanner = None

    def _scan_push_binding_accessor(self):  # type: () -> None
        assert self._scanner is not None, "parsing is done"
        self.parts.append(self._scan_binding_accessor())

    def _scan_binding_accessor(self):  # type: () -> BindingAccessorPart
        assert self._scanner is not None, "parsing is done"
        return BindingAccessorPart(scanner=self._scanner)

    _TRANSFORM_REGEX = re.compile(r"\w+")
    _TRANSFORM_PARAM_SEPARATOR_REGEX = re.compile(r",\s*")

    def _scan_push_transform(self):  # type: () -> None
        assert self._scanner is not None, "parsing is done"
        match = self._scanner.scan_regex(self._TRANSFORM_REGEX)
        if match:
            transform_name = match.group()
            transform = self._TRANSFORM_FUNCS.get(transform_name)
            if transform is None:
                raise ValueError("unknown transformation {!r}".format(transform_name))
            params = []
            if self._scanner.scan_string("("):
                params.append(self._scan_binding_accessor())
                for _ in self._scanner.scan_until(")"):
                    if self._scanner.scan_regex(self._TRANSFORM_PARAM_SEPARATOR_REGEX):
                        params.append(self._scan_binding_accessor())
            self.parts.append(BindingAccessorTransform(transform_name, transform, params))

    def resolve(self, **kwargs):  # type: (Any) -> Any
        """Given a context, resolve the expression and return the value."""
        value = None
        exc = None
        for part in self.parts:
            try:
                if isinstance(part, BindingAccessorPart):
                    value = part.resolve(**kwargs)
                elif isinstance(part, BindingAccessorTransform):
                    value = part.func(value, exc, part.params, **kwargs)
            except Exception:
                exc = sys.exc_info()
            else:
                exc = None
        if exc:
            reraise(*exc)
        return value

    @property
    def pretty_expression(self):  # type: () -> str
        """Format the expression in a human readable way."""
        expression = []
        for part in self.parts:
            if isinstance(part, BindingAccessorPart):
                expression.append(part.pretty_expression)
            elif isinstance(part, BindingAccessorTransform):
                expression.append(" | {}".format(part.name))
                if part.params:
                    expression.append("({})".format(", ".join(
                        [a.pretty_expression for a in part.params])))
        return "".join(expression)

    def __repr__(self):  # type: () -> str
        return "{}({!r})".format(self.__class__.__name__, self.pretty_expression)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#


class RequestPrefixListFilter(object):
    """ Matcher class for request path prefixes
    """

    def __init__(self, prefixes=None):
        if prefixes is None:
            prefixes = []
        self.reset(prefixes)

    def reset(self, prefixes):
        """ Reset the list of request path prefixes
        """
        self._prefixes = list(prefixes)

    def __iter__(self):
        return iter(self._prefixes)

    def match(self, path):
        """ Check if a request path is matched by a prefix

        Return the first matching prefix, or None if the request path does not
        start with any recorded prefix.
        """
        if not self._prefixes:
            return
        for prefix in self._prefixes:
            if path.startswith(prefix):
                return prefix

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, self._prefixes)

    def __str__(self):
        return ", ".join(self)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Various utils
"""
import datetime
import json
import sys
import types
from collections import deque
from inspect import isclass
from itertools import islice
from logging import getLogger
from operator import methodcaller

from ._vendors.ipaddress import (  # type: ignore
    ip_address as _ip_address,
    ip_network as _ip_network,
)

if sys.version_info >= (3, 5):
    from typing import (
        Any,
        Callable,
        Dict,
        Iterable,
        List,
        Mapping,
        Optional,
        Set,
        Tuple,
    )

    HAS_TYPING = True

    ConvertFunc = Callable[[Any], Optional[str]]

else:
    from collections import Iterable, Mapping

    HAS_TYPING = False

LOGGER = getLogger(__name__)

HAS_ASYNCIO = sys.version_info >= (3, 4)

NONE_TYPE = type(None)

ZERO_TD = datetime.timedelta(0)

if sys.version_info[0] < 3:
    ALL_STRING_CLASS = basestring  # noqa
    STRING_CLASS = str
    UNICODE_CLASS = unicode  # noqa

    def iterkeys(d, **kwargs):
        return d.iterkeys(**kwargs)

    def itervalues(d, **kwargs):
        return d.itervalues(**kwargs)

    def create_bound_method(func, obj):
        return types.MethodType(func, obj, obj.__class__)

    class UTCTZ(datetime.tzinfo):
        """Basic UTC timezone"""

        def utcoffset(self, dt):
            return ZERO_TD

        def tzname(self, dt):
            return "UTC"

        def dst(self, dt):
            return ZERO_TD

    UTC = UTCTZ()

else:
    ALL_STRING_CLASS = str
    STRING_CLASS = str
    UNICODE_CLASS = str

    def iterkeys(d, **kwargs):
        return d.keys(**kwargs)

    def itervalues(d, **kwargs):
        return d.values(**kwargs)

    create_bound_method = types.MethodType

    UTC = datetime.timezone.utc


def is_string(value):  # type: (Any) -> bool
    """ Check if a value is a valid string, compatible with python 2 and python 3

    >>> is_string('foo')
    True
    >>> is_string(u'✌')
    True
    >>> is_string(42)
    False
    >>> is_string(('abc',))
    False
    """
    return isinstance(value, ALL_STRING_CLASS)


def is_unicode(value):  # type: (Any) -> bool
    """ Check if a value is a valid unicode string, compatible with python 2 and python 3

    >>> is_unicode(u'foo')
    True
    >>> is_unicode(u'✌')
    True
    >>> is_unicode(b'foo')
    False
    >>> is_unicode(42)
    False
    >>> is_unicode(('abc',))
    False
    """
    return isinstance(value, UNICODE_CLASS)


def to_latin_1(value):  # type: (str) -> bytes
    """ Return the input string encoded in latin1 with replace mode for errors
    """
    return value.encode("latin-1", "replace")


def is_json_serializable(value):  # type: (Any) -> bool
    """ Check that a single value is json serializable
    """
    return isinstance(value, (ALL_STRING_CLASS, NONE_TYPE, bool, int, float))  # type: ignore


def to_unicode_safe(value):  # type: (Any) -> Optional[str]
    """ Safely convert a value to a unicode string.
    """
    if value is None:
        return value
    # If value is a byte string (string without encoding)
    # Try to decode it as unicode, this operation will
    # always succeed because non UTF-8 characters will
    # get replaced by the UTF-8 replacement character.
    if isinstance(value, bytes):
        return value.decode("utf-8", errors="replace")
    # If the value is not already a unicode string
    # Try to convert it to a string
    # by calling the standard Python __str__ method on
    # the value.
    elif not is_unicode(value):
        value = STRING_CLASS(value)
        # In Python 2.7, the returned value has no encoding
        if isinstance(value, bytes):
            return value.decode("utf-8", errors="replace")
    # Value is already a unicode string
    return value


###
# Raven configuration
###


def _raven_ignoring_handler(logger, *args, **kwargs):
    """ Ignore all logging messages from sqreen.* loggers, effectively
    disabling raven to log sqreen log messages as breadcrumbs
    """
    try:
        if logger.name.startswith("sqreen"):
            return True
    except Exception:
        LOGGER.warning("Error in raven ignore handler", exc_info=True)


def configure_raven_breadcrumbs():
    """ Configure raven breadcrumbs logging integration if raven is present
    """
    try:
        from raven import breadcrumbs  # type: ignore
    except ImportError:
        return

    # Register our logging handler to stop logging sqreen log messages
    # as breadcrumbs
    try:
        breadcrumbs.register_logging_handler(_raven_ignoring_handler)
    except Exception:
        LOGGER.warning("Error while configuring breadcrumbs", exc_info=True)


###
# NewRelics
###

def configure_newrelics_ignore_exception():
    """ Check if newrelics package is here, and disable Sqreen Exception
    """

    try:
        import newrelic.agent  # type: ignore
    except ImportError:
        return

    try:
        newrelic.agent.global_settings().error_collector.ignore_errors.extend(['sqreen.exceptions:RequestBlocked',
                                                                               'sqreen.exceptions:AttackBlocked',
                                                                               'sqreen.exceptions:ActionBlock',
                                                                               'sqreen.exceptions:ActionRedirect'])
    except Exception:
        LOGGER.warning("Error while configuring newrelics", exc_info=True)


###
# JSON Encoder
###


def qualified_class_name(obj):  # type: (Any) -> str
    """ Return the full qualified name of the class name of obj in form of
    `full_qualified_module.class_name`
    """
    if isclass(obj):
        instance_class = obj
    else:
        instance_class = obj.__class__

    return ".".join([instance_class.__module__, instance_class.__name__])


def django_user_conversion(obj):  # type: (Any) -> Optional[str]
    """ Convert a Django user either by returning USERNAME_FIELD or convert
    it to str.
    """
    if hasattr(obj, "USERNAME_FIELD"):
        return to_unicode_safe(getattr(obj, getattr(obj, "USERNAME_FIELD"), None))
    else:
        return UNICODE_CLASS(obj)


def psycopg_composable_conversion(obj, max_depth=64):  # type: (Any, int) -> str
    """ Best effort psycopg2 Composable string convertion.
    """
    if max_depth == 0:
        raise ValueError("Max depth reached while converting a psycopg2.sql.Composable")
    if hasattr(obj, "strings"):  # psycopg2.sql.Identifier
        # escaping is explained on https://www.postgresql.org/docs/current/sql-syntax-lexical.html#
        return ".".join(json.dumps(i).replace('\\"', '""') for i in obj.strings if isinstance(i, str))
    if hasattr(obj, "wrapped"):  # psycopg2.sql.Literal
        if isinstance(obj.wrapped, str):
            return "'{}'".format(obj.wrapped.replace("'", "''"))
        return repr(obj.wrapped)
    elif hasattr(obj, "__iter__"):  # psycopg2.sql.Composed
        return "".join(psycopg_composable_conversion(i, max_depth=max_depth - 1) for i in obj)
    return obj.as_string(None)


OBJECT_STRING_MAPPING = {
    "bson.objectid.ObjectId": UNICODE_CLASS,
    # Convert datetime to isoformat, compatible with Node Date()
    "datetime.datetime": methodcaller("isoformat"),
    "django.contrib.auth.models.AbstractUser": django_user_conversion,
    "os.PathLike": methodcaller("__fspath__"),
    "pathlib.PurePath": UNICODE_CLASS,
    "psycopg2.sql.Composable": psycopg_composable_conversion,
    "sqlalchemy.sql.elements.ClauseElement": UNICODE_CLASS,
    "sqreen._vendors.ipaddress.IPv4Address": UNICODE_CLASS,
    "sqreen._vendors.ipaddress.IPv6Address": UNICODE_CLASS,
}  # type: Dict[str, ConvertFunc]


def convert_to_string(obj, object_string_mapping=OBJECT_STRING_MAPPING):
    # type: (Any, Mapping[str, ConvertFunc]) -> Optional[str]
    """ Return a string representation for objects that are known to safely convert to string.
    """
    if type(obj) == type:
        instance_class = obj
    else:
        instance_class = obj.__class__

    # Manually do isinstance without needed to have a reference to the class
    for klass in instance_class.__mro__:
        qualified_name = qualified_class_name(klass)
        func = object_string_mapping.get(qualified_name)
        if func is not None:
            try:
                return func(obj)
            except Exception:
                LOGGER.warning("Error converting an instance of type %r", obj.__class__, exc_info=True)
    raise ValueError("cannot convert safely {} to string".format(instance_class))


class CustomJSONEncoder(json.JSONEncoder):
    MAPPING = OBJECT_STRING_MAPPING

    def default(self, obj):  # type: (Any) -> Optional[str]
        """ Return the repr of unkown objects
        """
        try:
            return convert_to_string(obj, object_string_mapping=self.MAPPING)
        except ValueError:
            # If we don't, or if we except, fallback on repr
            return repr(obj)


def ip_address(address):
    return _ip_address(UNICODE_CLASS(address))


def ip_network(address, strict=True):
    return _ip_network(UNICODE_CLASS(address), strict=strict)


def truncate_time(dt=None, round_to=60):  # type: (Optional[datetime.datetime], int) -> datetime.datetime
    """Return a datetime rounded to the previous given second interval."""
    if dt is None:
        dt = now()
    dt = dt.replace(microsecond=0)
    if dt.tzinfo is None:
        naive_dt = dt
    else:
        offset = dt.utcoffset()
        if offset is None:
            offset = ZERO_TD
        naive_dt = dt.replace(tzinfo=None) - offset
    seconds = (naive_dt - naive_dt.min).total_seconds()
    return dt - datetime.timedelta(seconds=seconds % round_to)


def flatten(iterable, max_iterations=1000):
    # type: (Iterable, int) -> Tuple[List, List]
    """Return the list of keys and values of iterable and nested iterables."""
    iteration = 0
    keys = []  # type: List[Any]
    values = []  # type: List[Any]
    remaining_iterables = deque([iterable], maxlen=max_iterations)
    seen_iterables = set()  # type: Set[int]

    while remaining_iterables:

        iteration += 1
        # If we have a very big or nested iterable, returns False.
        if iteration >= max_iterations:
            break

        iterable = remaining_iterables.popleft()
        id_iterable = id(iterable)
        # Protection against recursive objects
        if id_iterable in seen_iterables:
            continue
        seen_iterables.add(id_iterable)

        # If we get an iterable, add it to the list of remaining iterables.
        if isinstance(iterable, Mapping):
            keys.extend(islice(iterkeys(iterable), max_iterations))
            remaining_iterables.extend(islice(itervalues(iterable), max_iterations))
        elif isinstance(iterable, (list, tuple)):
            remaining_iterables.extend(islice(iter(iterable), max_iterations))
        else:
            values.append(iterable)

    return keys, values


def naive_dt_to_utc(dt):  # type: (datetime.datetime) -> datetime.datetime
    """Convert naive datetime to timezone aware datetime.
    By default, we consider all naive datetime to be in UTC.
    """
    if dt.tzinfo is None:
        return dt.replace(tzinfo=UTC)
    return dt


def now():  # type: () -> datetime.datetime
    """ Return the current UTC time.
    """
    return datetime.datetime.now(UTC)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
import sys

from ..._vendors.wrapt import FunctionWrapper
from .base import BaseMiddleware


class PyramidMiddleware(BaseMiddleware):

    def __call__(self, original):

        def wrapper(wrapped, instance, args, kwargs):
            """ Call the lifecycles methods with these arguments:
            Pyramid pre callbacks will receive these arguments:
            (None, request)
            Pyramid post callbacks will receive these arguments:
            (None, response)
            Pyramid failing callbacks will receive these arguments:
            (None, exception)
            """
            from pyramid.response import Response  # type: ignore

            request = args[0]
            self.strategy.before_hook_point()
            pre_args = (request,)
            self.execute_pre_callbacks(pre_args, record_attack=True)

            try:
                response = wrapped(*args, **kwargs)
            except Exception as e:
                if isinstance(e, Response):
                    self.execute_post_callbacks(e)
                else:
                    self.execute_failing_callbacks(sys.exc_info())
                raise

            return self.execute_post_callbacks(
                response, record_attack=True
            )

        return FunctionWrapper(original, wrapper)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Django hook strategy
"""

from logging import getLogger

import pkg_resources

from ..middlewares.django_middleware import DjangoMiddleware
from .framework import FrameworkStrategy

LOGGER = getLogger(__name__)


def int_or_betaversion(strversion):
    """ Convert a int as a str to either an int of 9999
    >>> int_or_betaversion("12")
    12
    >>> int_or_betaversion("2b1")
    9999
    """
    try:
        return int(strversion)
    except ValueError:
        return 9999


def _parse_version(version):
    """ Parse the django version

    >>> _parse_version("2.0")
    (2, 0)
    >>> _parse_version("2.2b1")
    (2, 9999)
    """
    return tuple(map(int_or_betaversion, version.split(".")[:2]))


def get_django_version():
    """Return the Django version pair (major, minor).

    Return None if no Django package can be found.
    """
    for package in pkg_resources.working_set:
        if package.project_name in ("Django", "django"):
            django_version = _parse_version(package.version)
            return django_version
    return None


DJANGO_VERSION = get_django_version()


def load_middleware_insert(original, middleware):
    def wrapped_load_middleware(self, *args, **kwargs):
        LOGGER.debug("Execute load_middleware_insert")

        # Load original middlewares.
        result = original(self, *args, **kwargs)

        if DJANGO_VERSION is None:
            LOGGER.warning(
                "Couldn't determine Django version, "
                "fallback on old-style middleware"
            )
            insert_middleware_v1(self)
        else:
            LOGGER.debug("Detected Django version %s", DJANGO_VERSION)
            if DJANGO_VERSION < (2, 0):
                insert_middleware_v1(self)
            else:
                insert_middleware_v2(self)

        return result

    def insert_middleware_v1(self):
        LOGGER.debug("Insert old-style Django middleware")

        # Insert Sqreen middleware.
        try:
            self._view_middleware.insert(0, middleware.process_view)
            self._response_middleware.append(middleware.process_response)
            self._exception_middleware.append(middleware.process_exception)
        except Exception:
            LOGGER.warning(
                "Error while inserting our middleware", exc_info=True
            )

    def insert_middleware_v2(self):
        LOGGER.debug("Insert new-style Django middleware")

        # Retrieve the original middleware chain.
        orig_mw_chain = self._middleware_chain

        # New middleware chain, including Sqreen. This function processes
        # responses, so Sqreen middleware is the last one.
        def mw_chain(request):
            response = orig_mw_chain(request)
            response = middleware.process_response(request, response)
            return response

        # Insert Sqreen middleware.
        try:
            self._view_middleware.insert(0, middleware.process_view)
            self._exception_middleware.append(middleware.process_exception)
            self._middleware_chain = mw_chain
        except Exception:
            LOGGER.warning(
                "Error while inserting our middleware", exc_info=True
            )

    return wrapped_load_middleware


class DjangoStrategy(FrameworkStrategy):
    """ Strategy for Django peripheric callbacks.

    It injects a custom DjangoFramework that calls callbacks for each
    lifecycle method
    """

    MODULE_NAME = "django.core.handlers.base"
    HOOK_CLASS = "BaseHandler"
    HOOK_METHOD = "load_middleware"

    def __init__(
        self,
        strategy_key,
        observation_queue,
        queue,
        import_hook,
        before_hook_point=None,
    ):
        super(DjangoStrategy, self).__init__(
            strategy_key,
            observation_queue,
            queue,
            import_hook,
            before_hook_point,
        )

        self.middleware = DjangoMiddleware(self, observation_queue, queue)
        self.wrapper = load_middleware_insert



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
import logging
from collections import defaultdict, deque
from datetime import timedelta
from itertools import groupby
from operator import itemgetter

from .events import RequestRecord
from .performance_metrics import RequestTrace, Trace
from .runner import MetricsEvent

LOGGER = logging.getLogger(__name__)

# Traces can be recorded out of a request, we keep the traces
# in the request recorder while waiting for the flush call. If there
# is no request for a while, the traces could pile up, so we limit their
# number and keep only the latest.
MAX_REQUEST_RECORDER_TRACES = 10000


class RequestRecorder:
    """Store observations related to a request."""

    def __init__(self):
        self.traces = deque(maxlen=MAX_REQUEST_RECORDER_TRACES)
        self.clear()

    def clear(self):
        """Clear all observations."""
        self.report = False
        self.payload_sections = set()
        self.observations = defaultdict(list)
        self.traces.clear()
        self.identity = None
        self.request_trace = None

    def trace(self, name=None):
        """Start a new trace from a request trace."""
        if self.request_trace is None:
            # There is not request at the moment, use a independant trace.
            return Trace(name, recorder=self)
        return self.request_trace.trace(name)

    def start_request(self, request, trace_request=False):
        """Start a request."""
        if trace_request:
            self.request_trace = RequestTrace(recorder=self)

    def end_request(self):
        """End the current request."""
        if self.request_trace is not None:
            self.request_trace.end()

    def record_trace(self, name, start_time, value, at):
        """Record a performance metric."""
        if name is not None:
            self.traces.append((name, at, None, value))

    @property
    def request_sq_time(self):
        """Total time spent in callbacks during a request."""
        if self.request_trace is not None:
            return self.request_trace.aggregated_duration_ms

    def observe(self, what, observation, payload_sections=None, report=True):
        """Record an observation."""
        if what == "sdk":
            name = observation[0]
            if name == "identify":
                self.identity = observation[2]
            elif self.identity and name == "track":
                identity = observation[3].get("user_identifiers")
                if identity and identity != self.identity:
                    LOGGER.warning(
                        "sqreen.identify and sqreen.track have been called "
                        "with different user_identifiers values"
                    )
        self.observations[what].append(observation)
        if payload_sections:
            self.payload_sections.update(payload_sections)
        if report:
            self.report = True

    def flush(self, request, response, payload_creator, queue, observation_queue):
        """Flush all observations."""
        if self.identity:
            # Tag unidentified track events with identity.
            for observation in self.observations["sdk"]:
                if observation[0] != "track":
                    continue
                track_options = observation[3]
                track_options.setdefault("user_identifiers", self.identity)

        self._put_metrics(queue, observation_queue)
        if self.report:
            payload = payload_creator.get_payload(
                request, response, self.payload_sections
            )
            # If request recorder must report the request, send also the
            # detailed observations.
            payload["observed"] = self.observations
            queue.put(RequestRecord(payload))
        self.clear()

    def _put_metrics(self, queue, observation_queue):
        if LOGGER.isEnabledFor(logging.DEBUG):
            it = groupby(self.traces, key=itemgetter(0))
            metrics = []
            for key, group in it:
                group = list(group)
                line = ""
                if key.startswith("sq."):
                    approx_date = \
                        group[0][1] - timedelta(seconds=group[0][3] / 1000.0)
                    line += str(approx_date)
                line += "\t{0}: ".format(key)
                line += " ".join("{0:.02f}".format(i[3]) for i in group)
                if len(group) > 1:
                    line += " ({:.02f} total)".format(sum(i[3] for i in group))
                metrics.append(line)
            if metrics:
                LOGGER.debug("Performance metrics:\n%s", "\n".join(metrics))

        # Send the performance metrics to observation queue for aggregation
        while self.traces:
            observation_queue.put(self.traces.popleft())

        # Pop the overtime callback observations so that they don't appear
        # in the request record when it is reported and only send the
        # first one for aggregation.
        overtime_callbacks = self.observations.pop("overtime_callbacks", [])
        if overtime_callbacks:
            observation_queue.put(overtime_callbacks[0])

        # Aggregate metric observations
        for observation in self.observations.pop("observations", []):
            observation_queue.put(observation)
        if observation_queue.half_full():
            queue.put(MetricsEvent)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
"""Middleware class for aiohttp."""

import sys
from asyncio import coroutine
from logging import getLogger

from ..._vendors.wrapt import FunctionWrapper
from ...exceptions import RequestBlocked
from ...runtime_storage import runtime
from .base import BaseMiddleware

LOGGER = getLogger(__name__)


def aiohttp_web_middleware(func):
    """Identical to aiohttp.web.middleware."""
    # This function was reimplemented in the agent to avoid importing aiohttp.
    func.__middleware_version__ = 1
    return func


class AioHTTPMiddleware(BaseMiddleware):
    """Middleware class for aiohttp."""

    # Decorator aiohttp.web.middleware is used internally by aiohttp to
    # distinguish between "regular" middlewares (decorated) and middleware
    # factories default.

    def __call__(self, original):
        """Wrap the original aiohttp middleware with callbacks."""
        from aiohttp import web  # type: ignore

        @aiohttp_web_middleware
        @coroutine
        def wrapper(wrapped, instance, args, kwargs):
            try:
                request = instance
                self.strategy.before_hook_point()
                self.execute_pre_callbacks((request,), record_attack=True)
                try:
                    response = yield from wrapped(*args, **kwargs)
                except web.HTTPException as exc:
                    # Each HTTP status code is a subclass of HTTPException.
                    # These exceptions are also a subclass of Response.
                    response = exc
                except Exception:
                    self.execute_failing_callbacks(sys.exc_info())
                    raise
                return self.execute_post_callbacks(
                    response, record_attack=True
                )
            except RequestBlocked:
                runtime.clear_request(self.queue, self.observation_queue)
                raise
            finally:
                runtime.clear_request()

        return FunctionWrapper(original, wrapper)

    # The decorator order is important here. The flag aiohttp.web.middleware is
    # lost on a static method, and aiohttp will consider this method to be a
    # middleware factory and raise an error if the middleware is enabled but not
    # hooked.

    @staticmethod
    @aiohttp_web_middleware
    @coroutine
    def handle(request, handler):
        """Instrumentation point, do nothing."""
        resp = yield from handler(request)
        return resp

    @staticmethod
    @coroutine
    def factory(app, handler):
        """Old style middleware factory, used in aiohttp 2.2."""

        # This function is a coroutine (direct call to another coroutine
        # function).
        def middleware_handler(request):
            return AioHTTPMiddleware.handle(request, handler)

        return middleware_handler



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Look for injected shell vars
"""
import os
from logging import getLogger

from ..utils import Mapping, is_string
from .regexp_rule import RegexpRule

LOGGER = getLogger(__name__)


class ShellEnvCB(RegexpRule):
    def pre(self, instance, args, kwargs, **options):
        """
        Examine environment variables to find shellshock injections.
        Works if env is passed in positional or keyword argument.
        """

        environ = kwargs.get("env")
        if not environ:
            # position based on
            # https://docs.python.org/3.6/library/subprocess.html#popen-constructor
            if len(args) >= 11:
                environ = args[10]

        if not environ:
            environ = os.environ

        if not isinstance(environ, (dict, Mapping)):
            return

        found, var = None, None
        for var, val in environ.items():
            if is_string(val):
                found = self.match_regexp(val)
                if found:
                    break

        if not found:
            return

        infos = {"variable_name": var, "variable_value": val, "found": found}
        self.record_attack(infos)

        return {"status": "raise", "data": found, "rule_name": self.rule_name}



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
import time
from datetime import datetime
from logging import getLogger
from operator import itemgetter

from .utils import UTC, now

DEFAULT_PERF_LEVEL = 0  # 0: disabled; 1: enabled
DEFAULT_PERF_PERIOD = 60
DEFAULT_PERF_BASE = 2.0
DEFAULT_PERF_UNIT = 0.1  # ms
DEFAULT_PERF_PCT_BASE = 1.3
DEFAULT_PERF_PCT_UNIT = 1.0  # %
DEFAULT_OVERTIME_PERIOD = 60

LOGGER = getLogger(__name__)


class PerformanceMetricsSettings:
    """ Performance metrics settings.
    """

    SETTINGS_FIELD_MAP = {
        'perf_level': 'level',
        'performance_metrics_period': 'period',
        'perf_base': 'base',
        'perf_unit': 'unit',
        'perf_pct_base': 'pct_base',
        'perf_pct_unit': 'pct_unit',
        'request_overtime_metric_period': 'overtime_period',
    }

    def __init__(self, level=DEFAULT_PERF_LEVEL,
                 period=DEFAULT_PERF_PERIOD,
                 base=DEFAULT_PERF_BASE,
                 unit=DEFAULT_PERF_UNIT,
                 pct_base=DEFAULT_PERF_PCT_BASE,
                 pct_unit=DEFAULT_PERF_PCT_UNIT,
                 overtime_period=DEFAULT_OVERTIME_PERIOD):
        self.level = level
        self.period = period
        self.base = base
        self.unit = unit
        self.pct_base = pct_base
        self.pct_unit = pct_unit
        self.overtime_period = overtime_period
        if self.enabled() and self.period == 0:
            LOGGER.warning("Setting performance period to default %d",
                           DEFAULT_PERF_PERIOD)
            self.period = DEFAULT_PERF_PERIOD

    @staticmethod
    def from_features(features):
        level = features.get('perf_level', DEFAULT_PERF_LEVEL)
        # old name, in Ruby not used for binned metrics:
        period = features.get('performance_metrics_period', DEFAULT_PERF_PERIOD)
        base = features.get('perf_base', DEFAULT_PERF_BASE)
        unit = features.get('perf_unit', DEFAULT_PERF_UNIT)
        pct_base = features.get('perf_pct_base', DEFAULT_PERF_PCT_BASE)
        pct_unit = features.get('perf_pct_unit', DEFAULT_PERF_PCT_UNIT)
        overtime_period = features.get('request_overtime_metric_period', DEFAULT_OVERTIME_PERIOD)

        return PerformanceMetricsSettings(level=level, period=period,
                                          base=base, unit=unit,
                                          pct_base=pct_base, pct_unit=pct_unit,
                                          overtime_period=overtime_period)

    def as_features(self):
        return {k: getattr(self, v) for k, v in self.SETTINGS_FIELD_MAP.items()}

    def enabled(self):
        return self.level > 0


class Trace(object):
    """Trace execution time."""

    def __init__(self, name=None, recorder=None):
        self.name = name
        self.recorder = recorder
        # TODO use a monotonic clock but there is none on Python < 3.3.
        # Maybe work around this with a native extension.
        self.start_time = time.time()
        self.duration = None

    @property
    def duration_ms(self):
        """Convert the duration from seconds to milliseconds."""
        if self.duration is not None:
            return self.duration * 1000

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        try:
            self.end()
        except Exception:
            LOGGER.error("exception while ending a performance trace", exc_info=1)

    def end(self, at=None):
        """End a performance trace."""
        if self.duration is not None:
            LOGGER.debug("A trace was ended multiple times: %r", self)
            return
        end_time = time.time()
        self.duration = end_time - (self.start_time or end_time)
        if self.duration < 0:
            self.duration = 0
        if self.recorder is not None:
            if at is None:
                at = datetime.fromtimestamp(end_time, tz=UTC)
            self.recorder.record_trace(self.name, self.start_time, self.duration_ms, at)

    def __repr__(self):
        return "<{} name={!r} duration_ms={:.4f}>".format(
            self.__class__.__name__,
            self.name,
            self.duration_ms
        )


class RequestTrace(Trace):
    """Trace request execution time for performance monitoring.

    Sub traces can be created for monitoring rule execution
    time. When the request is terminated, the total rule execution
    time is computed. Beware, a sub trace cannot survive this trace,
    meaning when the end method is called, all sub-traces are ended too.
    """

    def __init__(self, recorder=None):
        self._traces = []
        self._timeline = []
        super(RequestTrace, self).__init__("req", recorder)

    def trace(self, name=None):
        """Start a sub-trace."""
        trace = Trace(name, recorder=self)
        # Keep an history of all traces to compute total sub trace duration
        self._timeline.append((1, (trace.start_time - self.start_time) * 1000.0))
        self._traces.append(trace)
        return trace

    def _compute_aggregated_duration(self):
        # This is a modified version of the algorithm used to solve the
        # "minimum number of platforms required" problem.
        # It computes the overall time spent in sub-traces taking into
        # account the overlapping traces.
        counter = 0
        start_t = 0
        end_t = 0
        duration = 0
        # The timeline should be already in order, make sure it is
        self._timeline.sort(key=itemgetter(1))
        for op, t in self._timeline:
            # Start interval
            if counter == 0 and op == 1:
                start_t = t
            # End interval
            elif counter == 1 and op == -1:
                duration += t - start_t
                end_t = t
            elif op == -1:
                end_t = t
            counter += op
        # If some traces are not finished, use the last interval end
        if counter > 0:
            duration += end_t - start_t
        return duration if duration >= 0 else 0

    @property
    def aggregated_duration_ms(self):
        """Return total time spent in sub-traces."""
        return self._compute_aggregated_duration()

    def record_trace(self, name, start_time, duration, at):
        """Aggregate values from sub-traces."""
        # End of a sub-trace, keep it in the timeline to compute aggregated
        # duration
        self._timeline.append((-1, (start_time - self.start_time) * 1000.0 + duration))
        if self.recorder is not None:
            self.recorder.record_trace(name, start_time, duration, at)

    def end(self, at=None):
        """End the trace and all sub traces."""
        # Automatically terminate sub traces to assure consistency
        if at is None:
            at = now()
        for trace in self._traces:
            if trace.duration is None:
                trace.end(at=at)
        super(RequestTrace, self).end(at=at)
        if self.recorder is not None:
            aggregated_duration_ms = self._compute_aggregated_duration()
            self.recorder.record_trace("sq", None, aggregated_duration_ms, at)
            no_sq = self.duration_ms - aggregated_duration_ms
            if no_sq > 0:
                fraction = self.aggregated_duration_ms / no_sq
                self.recorder.record_trace("pct", None, 100.0 * fraction, at)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Script modifying the env to run our sitecustomize script before everyone else
"""
import argparse
import os
from os.path import dirname

from ..constants import INSTALLATION_URL


def where_sitecustomize():
    """ Returns the path of the sitecustomize package
    """
    current_file = os.path.split(__file__)[0]
    path = os.path.join(dirname(current_file), "sitecustomize")
    return path


def create_parser(parser_cls=argparse.ArgumentParser):
    parser = parser_cls(
        description=(
            "Launch a web application protected by the Sqreen agent.\n"
            "You can check {} for more information.".format(INSTALLATION_URL)
        )
    )
    parser.add_argument("command", help="command to run the application")
    parser.add_argument("arguments", nargs=argparse.REMAINDER, metavar="...")
    return parser


def prepend_pythonpath(pythonpath, environ):
    patched_environ = dict(environ)
    current_pythonpath = environ.get("PYTHONPATH")
    if current_pythonpath is None:
        current_pythonpath = pythonpath
    else:
        current_pythonpath = os.pathsep.join([pythonpath, current_pythonpath])
    patched_environ["PYTHONPATH"] = current_pythonpath
    return patched_environ


def protect(args=None):
    """ Call the passed binary passed in argument with the right PYTHONPATH
    """
    parser = create_parser()
    args = parser.parse_args(args)
    environ = prepend_pythonpath(where_sitecustomize(), os.environ)
    os.execvpe(args.command, [args.command] + args.arguments, environ)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Collect and aggregate metrics
"""
import math
from collections import Counter, defaultdict, namedtuple
from datetime import timedelta
from logging import getLogger

from .exceptions import SqreenException
from .utils import truncate_time

LOGGER = getLogger(__name__)


class BaseAggregator(object):
    """ Base classes for aggregators.

    It contains the methods to update a period with new data and finalize the
    data just before uploading.
    Aggregators don't store any state, it creates, updates, and finizalises
    an accumulator that is passed by the MetricsStore.
    """

    name = "Base"

    @staticmethod
    def default_value_factory():
        return None

    @classmethod
    def initialize(cls, **options):
        """ Initialize the accumulator.
        """
        return defaultdict(cls.default_value_factory)

    @staticmethod
    def is_empty(accumulator):
        return len(accumulator) == 0

    @classmethod
    def update(cls, key, value, accumulator):
        """ Update the accumulator for the period. For updating only the key,
        override update_key instead.
        """
        accumulator[key] = cls.update_key(value, accumulator[key])
        return accumulator

    @staticmethod
    def update_key(new_value, cur_value):
        """ Update the data for a key in the period. For updating the whole data,
        override update instead.
        """
        raise NotImplementedError("")

    @classmethod
    def finalize(cls, accumulator):
        """ Finalize the period, by default return the data. Could be override
        to compute a ration or sum some elements.
        """
        return accumulator


class CollectAggregator(BaseAggregator):
    """ Simple aggregator that stores all the observations points,
    used in tests mostly
    """

    name = "Collect"
    default_value_factory = list

    @staticmethod
    def update_key(new_value, cur_value):
        cur_value.append(new_value)
        return cur_value


class SumAggregator(BaseAggregator):

    name = "Sum"
    default_value_factory = int

    @staticmethod
    def update_key(new_value, cur_value):
        return cur_value + new_value


class ExtensionPerformanceMetricAggregator(BaseAggregator):

    name = "ExtensionPerformance"

    @classmethod
    def initialize(cls, **options):
        return {'v': Counter()}

    @staticmethod
    def is_empty(accumulator):
        return len(accumulator["v"]) == 0

    @classmethod
    def update(cls, key, value, accumulator):
        """ Ignore key
        """
        accumulator['u'] = value['u']
        accumulator['b'] = value['b']
        accumulator['v'].update(value['v'])
        return accumulator


class AverageAggregator(BaseAggregator):

    name = "Average"

    @staticmethod
    def default_value_factory():
        return {"sum": 0, "count": 0}

    @staticmethod
    def update_key(new_value, cur_value):
        cur_value["sum"] += new_value
        cur_value["count"] += 1
        return cur_value

    @classmethod
    def finalize(cls, accumulator):
        final_data = {}

        for key, value in accumulator.items():
            final_data[key] = value["sum"] / float(value["count"])

        return final_data


BinningAccumulator = namedtuple("BinningAccumulator", [
    "unit", "base", "inv_log_base", "sub_parcel", "values"
])


class BinningAggregator(BaseAggregator):
    """Sample numbers into logarithmic bins."""

    name = "Binning"

    @classmethod
    def initialize(cls, **options):
        unit = options.get("unit")
        base = options.get("base")
        if unit is None or not unit > 0.0:
            raise InvalidOptionAggregator("unit must be a positive value")
        if base is None or not base > 1.0:
            raise InvalidOptionAggregator("base must be greater than 1.0")
        log_base = math.log(base)
        log_unit = math.log(unit)
        return BinningAccumulator(unit, base, 1 / log_base, log_unit / log_base, Counter())

    @staticmethod
    def is_empty(accumulator):
        return len(accumulator.values) == 0

    @classmethod
    def update(cls, key, value, accumulator):
        # key is ignored
        if value < accumulator.unit:
            value_bin = 1
        else:
            value_bin = 2 + math.floor(math.log(value) * accumulator.inv_log_base
                                       - accumulator.sub_parcel)
        accumulator.values[value_bin] += 1
        accumulator.values["max"] = max(accumulator.values["max"], value)
        return accumulator

    @classmethod
    def finalize(cls, accumulator):
        return {
            "u": accumulator.unit,
            "b": accumulator.base,
            "v": dict(accumulator.values.items()),
        }


class InvalidOptionAggregator(SqreenException):
    """ Exception raised when an invalid option was passed to an aggregator.
    """


class UnknownAggregator(SqreenException):
    """ Exception raised when trying to register a metric with an unknown
    aggregation kind
    """


class AlreadyRegisteredMetric(SqreenException):
    """ Exception raised when trying to register twice the same metric
    name.
    """


class AlreadyRegisteredAggregator(SqreenException):
    """ Exception raised when trying to register twice the same aggregator
    name.
    """


class UnregisteredMetric(SqreenException):
    """ Exception raised when trying to update an unregistered metric.
    """


PRODUCTION_AGGREGATORS = [SumAggregator(),
                          AverageAggregator(),
                          ExtensionPerformanceMetricAggregator(),
                          BinningAggregator()]


class MetricsStore(object):
    """ Store the dict of currently monitored metrics.

    For each metric, store a dict containing:
    - The start time of monitored period.
    - The maximum period time.
    - The kind of aggregator.
    - The aggregated data, the value is managed by the aggregator directly.

    Store also the list of available aggregator indexed by a kind.

    When periods are finished, store them to be retrieved for pushing in
    the store attribute.
    """

    def __init__(self):
        self.store = []
        self.metrics = {}
        self.aggregators = {}

    def register_metric(self, name, kind, period, **options):
        """ Register a new metric
        """
        LOGGER.debug("Register metric '%s'", name)

        if name in self.metrics:
            existing_metric = self.metrics[name]
            saved_metric_kind = existing_metric["kind"]
            if saved_metric_kind != kind:
                err_msg = (
                    "Metric '{}' has already been registered with kind {}"
                )
                raise AlreadyRegisteredMetric(
                    err_msg.format(name, saved_metric_kind)
                )

            # Update the period
            existing_metric["period"] = period
            # Reset the metric if the options have changed
            if options != existing_metric["options"]:
                existing_metric["options"] = options
                self.metrics[name] = self.reset_metric(existing_metric)
        else:
            if kind not in self.aggregators:
                raise UnknownAggregator(
                    "Unknown aggregation kind: {}".format(kind)
                )
            metric = self._new_metric(kind, period, options)
            self.metrics[name] = self.reset_metric(metric)

    def unregister_metric(self, name):
        """ Unregister a metric
        """
        LOGGER.debug("Unregister metric %r", name)
        self.metrics.pop(name, None)

    def ensure_ext_perf_metric(self, name, period):
        if name in self.metrics:
            return

        self.register_metric(name, 'ExtensionPerformance', period)

    @staticmethod
    def _new_metric(kind, period, options):
        """ Return a dict for an empty metric period
        """
        return {
            "kind": kind,
            "period": period,
            "observation": None,
            "start": None,
            "options": options,
        }

    def register_aggregator(self, name, aggregator_function):
        """ Register a new aggregator under the name passed in input
        """
        if name in self.aggregators:
            msg = "Aggregator '{}' has already been registered to: {}"
            raise AlreadyRegisteredAggregator(
                msg.format(name, self.aggregators[name])
            )
        self.aggregators[name] = aggregator_function

    def register_production_aggregators(self):
        """ Register production aggregators
        """
        for aggregator in PRODUCTION_AGGREGATORS:
            self.register_aggregator(aggregator.name, aggregator)

    def register_default_metrics(self):
        """ Register production default metrics
        """

        self.register_metric("sqreen_call_counts", "Sum", 60)
        self.register_metric("whitelisted", "Sum", 60)
        self.register_metric("request_overtime", "Sum", 60)

    def update(self, name, at, key, value):
        """ Logic behind metric updating.

        Check if the start time is set for the metric period.
        Check if the metric period has expired, if so save it and create
        a blank metric period.
        Then call the aggregator to compute the new data
        """
        try:
            metric = self.metrics[name]
        except KeyError:
            raise UnregisteredMetric("Unknown metric {}".format(name))

        # Check if the metric should be published or not
        metric = self.check_metric_period_over(metric, name, at, False)
        # Update start time if not set already (registered but never updated)
        if metric["start"] is None:
            metric["start"] = truncate_time(at, round_to=metric["period"])
        self._update_metric(metric, key, value)

    def _update_metric(self, metric, key, value):
        """ Actual method that call the aggregator to compute the new data.
        """
        aggregator = self.aggregators[metric["kind"]]

        # Compute the new value
        new_data = aggregator.update(key, value, metric["observation"])
        metric["observation"] = new_data

    def check_metric_period_over(self, metric, name, at, force_finalize=True):
        """ Check a single metric to see if its period is over, if so
        finalize it and returns the new one.
        If force_finalize if False, check that the metric period is over first.
        """
        if metric["start"] is None:
            return metric

        period_over = (
            metric["start"] + timedelta(seconds=metric["period"]) < at
        )
        if force_finalize or period_over:
            return self.finalize_period(metric, name, at)

        return metric

    def check_all_metrics_period_over(self, at, force_finalize=True):
        """ Check all registered metrics to see if their period are over, if so
        finalize them.
        If force_finalize if False, check that the metric period is over first.
        """
        for metric_name, metric in self.metrics.items():
            self.check_metric_period_over(
                metric, metric_name, at, force_finalize
            )

    def finalize_period(self, metric, name, at):
        """ Finalize a metric period. For each registered metric, call the
        finalize method on correspondent aggregator and instantiate a new blank
        metric period.

        Called if either the period time was crossed or can forced on logout.
        """
        aggregator = self.aggregators[metric["kind"]]
        # If no data has been gathered
        if not aggregator.is_empty(metric["observation"]):
            # Retrieve the current period
            # Call the finalize method on the aggregator

            finished = dict(metric)
            finished["name"] = name
            finished["observation"] = aggregator.finalize(
                metric["observation"]
            )
            usual_finish = metric["start"] + timedelta(seconds=metric["period"])
            finished["finish"] = at if at < usual_finish else usual_finish
            self.store.append(finished)

        # Reset the period
        metric = self.reset_metric(metric, at)
        self.metrics[name] = metric
        return metric

    def reset_metric(self, metric, at=None):
        if at is not None:
            at = truncate_time(at, round_to=metric["period"])
        aggregator = self.aggregators[metric["kind"]]
        metric["observation"] = aggregator.initialize(**metric["options"])
        metric["start"] = at
        return metric

    def get_data_to_publish(self, at, force_finalize=True):
        """ Return the list of finished periods, reset the list of
        finished periods after.
        """
        self.check_all_metrics_period_over(at, force_finalize)

        finished_periods = self.store
        self.store = []
        return finished_periods

    ###
    # Helpers for debug and tests
    ###

    def list_metrics(self):
        """ Return the list of registered metrics
        """
        # The list must be a copy so that we can still register / unregister metrics.
        return list(self.metrics.keys())

    def is_registered(self, name):
        """ Return True if the metric was already registered.
        """
        return name in self.metrics

    def get_metric_kind(self, name):
        """ Return the kind of a given metric
        """
        return self.metrics[name]["kind"]

    def get_metric_period(self, name):
        """ Return the period of a given metric
        """
        return self.metrics[name]["period"]

    def get_metric_start(self, name):
        """ Return the start time for a given metric
        """
        return self.metrics[name]["start"]

    def get_metric_aggregate(self, name):
        """ Return the current aggregated data for a given metric
        """
        return self.metrics[name]["observation"]



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" High-level interaction with sqreen API
"""
import logging

from ._vendors.urllib3.exceptions import (  # type: ignore
    HTTPError,
    MaxRetryError,
)
from .exceptions import SqreenException
from .http_client import InvalidStatusCodeResponse, StatusFailedResponse
from .remote_exception import RemoteException

LOGGER = logging.getLogger(__name__)


class InvalidToken(Exception):
    """ Exception raise when a login fails because of the token value
    """


class InvalidSession(Exception):
    """ Exception raise when the session is not valid anymore.
    """


class Session(object):
    """ Class responsible for collection date and interacting with the sqreen API
    """

    def __init__(self, connection, api_key, app_name=None):
        self.connection = connection
        self.api_key = api_key
        self.app_name = app_name
        self.session_token = None

    def login(self, runtime_infos, retries=None):
        """ Login to the backend
        """

        if not retries:
            retries = self.connection.RETRY

        # We want repeat this logic for as long as necessary, but want to avoid a recursive call
        try:
            result = self.connection.post(
                "v1/app-login", runtime_infos, headers=self._api_headers(),
                retries=retries
            )
        except InvalidStatusCodeResponse as exc:
            LOGGER.error(
                "Cannot login. Token may be invalid: %s", self.api_key
            )
            LOGGER.error("Invalid response: %s", exc.response_data)
            if exc.status in (401, 403):
                raise InvalidToken
            raise
        except StatusFailedResponse as exc:
            LOGGER.error(
                "Cannot login. Token may be invalid: %s", self.api_key
            )
            LOGGER.error("Invalid response: %s", exc.response)
            raise InvalidToken
        except (SqreenException, HTTPError) as exc:
            LOGGER.error(
                "Cannot login. back.sqreen.io appears to be unavailable (%s).", type(exc).__name__
            )
            raise HTTPError

        LOGGER.debug("Received session_id %s", result["session_id"])
        self.session_token = result["session_id"]

        return result

    def is_connected(self):
        """ Return a boolean indicating if a successfull login was made
        """
        return self.session_token is not None

    def _headers(self):
        """Return session headers used for authentication."""
        return {"x-session-key": self.session_token}

    def _api_headers(self):
        """Return API headers."""
        if self.app_name:
            return {
                "x-api-key": self.api_key,
                "x-app-name": self.app_name
            }
        else:
            return {
                "x-api-key": self.api_key
            }

    def perform_network(self, net_operation, url, retries, payload=None, headers=None, last_resort=False):
        """ Perform a network operation safely
        """
        try:
            if headers is None:
                headers = self._headers()

            if payload is None:
                return net_operation(url, headers=headers, retries=retries)
            else:
                return net_operation(url, payload, headers=headers, retries=retries)

        except MaxRetryError:
            LOGGER.info(
                "Request to %s expired after %d retries, moving on", url, retries.total
            )

        except InvalidStatusCodeResponse as exc:
            if exc.status in (401, 403):
                raise InvalidSession
            LOGGER.info("Unexpected HTTP status code: %r", exc.status)

        except (SqreenException, HTTPError) as exc:
            LOGGER.info(
                "Couldn't connect to %s due to exception %s", url, type(exc).__name__
            )
            # Prevent recursive exception
            if not last_resort:
                self.post_sqreen_exception(RemoteException.from_exc_info().to_dict(), last_resort=True)

    def _get(self, url, retries=None):
        """ Call connection.get with right headers
        """
        return self.perform_network(self.connection.get, url=url, retries=retries)

    def _post(self, url, data, retries=None, last_resort=False):
        """Call connection.post with session headers."""
        return self.perform_network(self.connection.post, payload=data, url=url, retries=retries, last_resort=last_resort)

    def _post_api(self, url, data, retries=None):
        """Call connection.post with API headers."""
        return self.perform_network(self.connection.post, payload=data, url=url, headers=self._api_headers(), retries=retries)

    def logout(self):
        """ Logout current instance in the backend
        """
        return self._get("v0/app-logout", retries=self.connection.RETRY_ONCE)

    def heartbeat(self, payload):
        """ Tell the backend that the instance is still up, send latests command
        result, latest metrics and retrieve latest commands
        """
        return self._post(
            "v1/app-beat", payload, retries=self.connection.RETRY_LONG
        )

    def post_attack(self, attack):
        """ Report an attack on the backend
        """
        LOGGER.debug("Post attack %s", attack)
        return self._post(
            "v0/attack", attack, retries=self.connection.RETRY_LONG
        )

    def post_sqreen_exception(self, exception, last_resort=False):
        """Report a Sqreen exception happening at agent level."""
        return self._post(
            "v0/sqreen_exception", exception, retries=self.connection.RETRY_LONG,
            last_resort=last_resort
        )

    def post_app_sqreen_exception(self, exception):
        """Post a sqreen exception happening at application level."""
        return self._post_api(
            "v0/app_sqreen_exception", exception, retries=self.connection.RETRY_ONCE
        )

    def post_metrics(self, metrics):
        """ Post metrics aggregates to the backend
        """
        # Don't send empty metrics payload
        if len(metrics) < 1:
            return

        data = {"metrics": metrics}

        return self._post(
            "v0/metrics", data, retries=self.connection.RETRY_ONCE
        )

    def post_request_record(self, request_record):
        LOGGER.debug("Post request record %r", request_record)
        return self._post(
            "v0/request_record",
            request_record,
            retries=self.connection.RETRY_LONG
        )

    def get_rulespack(self):
        """ Retrieve rulespack from backend
        """
        return self._get("v0/rulespack", retries=self.connection.RETRY_LONG)

    def post_batch(self, batch, send_resiliently=True):
        """ Post a batch to the backend
        """
        LOGGER.debug("Post batch of size %d", len(batch))

        if send_resiliently:
            retry_policy = self.connection.RETRY_LONG
        else:
            retry_policy = self.connection.RETRY_ONCE

        return self._post(
            "v0/batch", {"batch": batch}, retries=retry_policy
        )

    def post_bundle(self, runtime_infos):
        data = {
            "bundle_signature": runtime_infos["bundle_signature"],
            "dependencies": runtime_infos["various_infos"]["dependencies"],
        }
        return self._post(
            "v0/bundle", data, retries=self.connection.RETRY_LONG
        )

    def get_actionspack(self):
        """Retrieve actions from backend."""
        return self._get("v0/actionspack", retries=self.connection.RETRY_LONG)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Pools context resources
"""
import logging
from threading import Lock
from weakref import WeakKeyDictionary

from py_mini_racer.py_mini_racer import MiniRacer  # type: ignore

from ..runner import BaseRunner
from ..utils import HAS_TYPING

if HAS_TYPING:
    from typing import MutableMapping

LOGGER = logging.getLogger(__name__)


class JSContextPool:
    def __init__(self):
        self.lock = Lock()
        self.contexts = []  # needed?
        self.total = 0

    def with_context(self, func):
        context = self._fetch_context()
        try:
            return func(context)
        finally:
            self._give_back_context(context)

    def _fetch_context(self):
        with self.lock:
            if not self.contexts:  # empty
                self.total += 1
                LOGGER.info("Creating new V8 context. "
                            "Count for this pool is %d", self.total)
                return DefinitionTrackingMiniRacer()
            else:
                return self.contexts.pop()

    def _give_back_context(self, context):
        context.garbage_collect()
        if context.gc_load > 30:
            if context.gc_threshold_in_bytes == DEFAULT_GC_THRESHOLD:
                context.gc_threshold_in_bytes *= 2
                context.set_soft_memory_limit(context.gc_threshold_in_bytes)
                LOGGER.warning("Context %s had too many close garbage "
                               "collections; doubling the threshold to %d bytes",
                               context, context.gc_threshold_in_bytes)
                context.gc_load = 0
            else:
                LOGGER.warning("Context %s had too many close garbage "
                               "collections; discarding it", context)
                del context
                return

        with self.lock:
            self.contexts.append(context)

    class DummyRunner(BaseRunner):
        pass

    _runner_pool_map = WeakKeyDictionary()  # type: MutableMapping[BaseRunner, JSContextPool]
    _dummy_runner = DummyRunner()

    @classmethod
    def fetch_for_runner(cls, runner):
        if runner is None:
            runner = cls._dummy_runner

        if runner in cls._runner_pool_map:
            return cls._runner_pool_map[runner]
        else:
            new_pool = cls()
            cls._runner_pool_map[runner] = new_pool
            return new_pool


DEFAULT_GC_THRESHOLD = 15000000  # 15 MB


class DefinitionTrackingMiniRacer(MiniRacer):

    def __init__(self, *args, **kwargs):
        super(DefinitionTrackingMiniRacer, self).__init__(*args, **kwargs)
        self.code_ids = set()
        self.failed_code_ids = set()
        self.gc_load = 0
        self.gc_threshold_in_bytes = DEFAULT_GC_THRESHOLD
        # compatibility with older versions of PyMiniRacer
        if not hasattr(self, "set_soft_memory_limit"):
            self.set_soft_memory_limit = self.legacy_set_soft_memory_limit
            self.was_soft_memory_limit_reached = self.legacy_was_soft_memory_limit_reached
        self.set_soft_memory_limit(self.gc_threshold_in_bytes)

    def add_code(self, code_id, code):
        try:
            LOGGER.debug("Adding code with md5 %s to %s", code_id, self)
            self.eval('(function() {{ {} }})()'.format(code))
            self._transf_global_funcs(code_id)
            self.code_ids.add(code_id)
        except Exception:
            self.failed_code_ids.add(code_id)
            raise

    def _transf_global_funcs(self, code_id):
        snippet = '''
            if (typeof sqreen_data === 'undefined') {{
              sqreen_data = {{}};
            }}

            group = {{}};
            Object.keys(this).forEach(name => {{
              if (typeof this[name] === "function") {{
                group[name] = this[name];
                this[name] = undefined;
              }}
            }});
            sqreen_data['{}'] = group
            delete group;
        '''.format(code_id)
        self.eval(snippet)

    def legacy_set_soft_memory_limit(self, value):
        pass

    def legacy_was_soft_memory_limit_reached(self):
        stats = self.heap_stats()
        return stats['total_heap_size'] > self.gc_threshold_in_bytes

    def garbage_collect(self):
        # garbage collections max 1 in every 4 calls (avg)
        if self.was_soft_memory_limit_reached():
            LOGGER.debug("Doing low memory notification for context")
            self.low_memory_notification()
            self.gc_load += 4
        else:
            self.gc_load = max(0, self.gc_load - 1)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Various constants
"""

LIFECYCLE_METHODS = {"PRE": "pre", "POST": "post", "FAILING": "failing"}

ACTIONS = {
    "RAISE": "raise",
    "ACTION_BLOCK": "action_block",
    "ACTION_REDIRECT": "action_redirect",
    "OVERRIDE": "override",
    "RETRY": "retry",
    "MODIFY_ARGS": "modify_args",
}


VALID_ACTIONS_PER_LIFECYCLE = {
    LIFECYCLE_METHODS["PRE"]: [
        ACTIONS["RAISE"],
        ACTIONS["ACTION_BLOCK"],
        ACTIONS["ACTION_REDIRECT"],
        ACTIONS["OVERRIDE"],
        ACTIONS["MODIFY_ARGS"],
    ],
    LIFECYCLE_METHODS["FAILING"]: [
        ACTIONS["RAISE"],
        ACTIONS["RETRY"],
        ACTIONS["OVERRIDE"],
    ],
    LIFECYCLE_METHODS["POST"]: [ACTIONS["RAISE"], ACTIONS["OVERRIDE"]],
}


BACKEND_URL = "https://back.sqreen.io"
CHANGELOG_URL = "https://docs.sqreen.io/sqreen-for-python/python-changelog/"
COMPATIBILITY_URL = (
    "https://docs.sqreen.io/sqreen-for-python/python-agent-compatibility/"
)
INGESTION_BACKEND_URL = "https://ingestion.sqreen.com/"
INSTALLATION_URL = (
    "https://docs.sqreen.io/sqreen-for-python/installing-the-python-agent/"
)
STATUS_URL = "http://status.sqreen.io/"
TERMS_URL = "https://www.sqreen.io/terms.html"



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Callback for tracing anything
"""
import logging

from ..binding_accessor import BindingAccessor
from ..rules import RuleCallback
from ..samplers import ScopedTracingSampler

LOGGER = logging.getLogger(__name__)


class Trace(RuleCallback):

    def __init__(self, *args, **kwargs):
        super(Trace, self).__init__(*args, **kwargs)
        self.payload_template = {}
        options = self.data.get("options", {})
        self.scope = options.get("scope")
        sampler_definition = options.get("sampler_definition")
        self.sampler = ScopedTracingSampler(sampler_definition) \
            if sampler_definition is not None else None
        for key, template in options.get("payload_template", {}).items():
            self.payload_template[key] = BindingAccessor(template)

    def pre(self, instance, args, kwargs, **options):
        binding_eval_args = {
            "request": self.storage.get_current_request(),
            "response": self.storage.get_current_response(),
            "inst": instance,
            "args": self.storage.get_current_args(args),
            "kwargs": kwargs,
            "data": self.data,
            "extra": options,
        }
        payload = {}
        try:
            for key, ba in self.payload_template.items():
                payload[key] = ba.resolve(**binding_eval_args)
        except Exception:
            LOGGER.debug("ignore this trace as the payload template cannot be filled", exc_info=True)
        else:
            if self.sampler is None or self.sampler.should_sample(scope=self.scope, payload=payload):
                # if there is a callback level sampler, use it then delegates the decision to the agent
                # to send or not the trace
                self.record_trace(scope=self.scope, payload=payload)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
import os
import sys


def detect_debug_mode():
    """ Detect if any supported and already loaded framework is in debug mode.

    Not all frameworks support accessing this information globally. In this
    case, the "is_debug" method is exposed on the request wrapper.
    """
    modes = []

    # Django
    django_conf = sys.modules.get("django.conf", None)
    if django_conf is not None and hasattr(django_conf, "settings"):
        modes.append(getattr(django_conf.settings, "DEBUG", False))

    # Pyramid
    modes.append(bool(os.environ.get("PYTHON_RELOADER_SHOULD_RUN", False)))

    # asyncio
    asyncio_module = sys.modules.get("asyncio", None)
    if asyncio_module is not None:
        try:
            modes.append(asyncio_module.get_event_loop().get_debug())
        except Exception:
            pass

    return any(modes)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Flask specific WSGI HTTP Request / Response stuff
"""

import sys
from logging import getLogger
from traceback import format_stack

from ..utils import to_unicode_safe
from .base import BaseResponse
from .wsgi import BaseWSGIRequest

if sys.version_info[0] >= 3:
    import urllib.parse as urlparse
else:
    import urlparse


LOGGER = getLogger(__name__)


class FlaskRequest(BaseWSGIRequest):

    def __init__(self, request, storage=None):
        super(FlaskRequest, self).__init__(storage=storage)
        self.request = request

        # Cache for params
        self._query_params = None
        self._form_params = None
        self._cookies_params = None
        self._json_params = None

    def preload_data(self):
        """ Some Flask properties consume the stream without keeping
        a cache of the input data. We must call preload_data first,
        otherwise get_json will eventually fail.
        """
        if hasattr(self.request, "_cached_data"):
            return True
        length = self.content_length
        # Safety, do not preload data if the body is too heavy
        # XXX: this should be a parameter.
        if length is None or length > (1024 * 1024):  # 1MB
            return False
        try:
            # Fill request._cached_data
            self.request.get_data()
            if self.is_flask_api:
                # Flask-API redefines several properties such as data and form
                # but they all depend on the _stream attribute. We reset it
                # in order to allow further input processing by the framework.
                self.request._stream = self.request._get_stream_for_parsing()
        except Exception:
            LOGGER.debug("couldn't get request.data from the framework",
                         exc_info=True)
            return False
        else:
            return True

    @property
    def query_params(self):
        if self._query_params is None:
            try:
                # Convert flask QueryDict to a normal dict with values as list
                self._query_params = dict(self.request.args.lists())
            except Exception:
                LOGGER.debug("couldn't get request.args from the framework",
                             exc_info=True)
                self._query_params = super(FlaskRequest, self).query_params
        return self._query_params

    @property
    def body(self):
        """ Get the HTTP input body from the framework cache or the
        WSGI input preview.
        """
        try:
            if self.preload_data():
                return self.request.get_data()
        except Exception:
            LOGGER.debug("couldn't get request.data from the framework",
                         exc_info=True)
        return super(FlaskRequest, self).body

    @property
    def form_params(self):
        if self._form_params is None:
            try:
                if self.preload_data():
                    self._form_params = dict(self.request.form.lists())
                    return self._form_params
            except Exception:
                LOGGER.debug("couldn't get request.form from framework",
                             exc_info=True)
            self._form_params = super(FlaskRequest, self).form_params
        return self._form_params

    @property
    def cookies_params(self):
        if self._cookies_params is None:
            try:
                self._cookies_params = dict(self.request.cookies)
            except Exception:
                LOGGER.debug("couldn't get request.cookies from framework",
                             exc_info=True)
                self._cookies_params = super(FlaskRequest, self).cookies_params
        return self._cookies_params

    @property
    def json_params(self):
        if self._json_params is None:
            try:
                if self.preload_data():
                    self._json_params = self.request.get_json(silent=True)
                    return self._json_params
            except Exception:
                LOGGER.debug("couldn't get request.json from the framework",
                             exc_info=True)
            self._json_params = super(FlaskRequest, self).json_params
        return self._json_params

    @property
    def remote_addr(self):
        """Remote IP address."""
        return to_unicode_safe(self.get_raw_header("REMOTE_ADDR"))

    @property
    def hostname(self):
        try:
            url = self.request.url_root
            return urlparse.urlparse(url).netloc
        except Exception:
            LOGGER.debug("couldn't get request.get_host from the framework",
                         exc_info=True)
            return None

    @property
    def method(self):
        return to_unicode_safe(self.request.method)

    @property
    def referer(self):
        return self.request.referrer

    @property
    def client_user_agent(self):
        return to_unicode_safe(self.request.user_agent.string)

    @property
    def route(self):
        """Request route."""
        url_rule = getattr(self.request, "url_rule", None)
        # If a custom rule_class is used in the app, the rule attribute might
        # be missing.
        return getattr(url_rule, "rule", None)

    @property
    def path(self):
        return self.request.path

    @property
    def scheme(self):
        return to_unicode_safe(self.request.scheme)

    @property
    def server_port(self):
        return to_unicode_safe(self.get_raw_header("SERVER_PORT"))

    @property
    def remote_port(self):
        return to_unicode_safe(self.get_raw_header("REMOTE_PORT"))

    @property
    def raw_headers(self):
        return self.request.environ

    @property
    def caller(self):
        return format_stack()

    @property
    def view_params(self):
        return self.request.view_args

    @property
    def is_flask_api(self):
        """ Detect a Flask API request. """
        return hasattr(self.request, "empty_data_class") \
            and "flask_api" in sys.modules

    @staticmethod
    def is_debug():
        try:
            from flask import current_app

            return bool(current_app.debug)
        except Exception:
            LOGGER.debug("couldn't get current_app.debug", exc_info=True)
            return False


class FlaskResponse(BaseResponse):

    def __init__(self, response):
        self.response = response

    @property
    def status_code(self):
        return self.response.status_code

    @property
    def content_type(self):
        return self.response.headers.get("Content-Type")

    @property
    def content_length(self):
        try:
            return int(self.response.headers.get("Content-Length"))
        except (ValueError, TypeError):
            return None



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Sqreen Python agent thread composition
"""
import atexit
import json
import os
import random
import sys
import threading
import time
import traceback
from copy import copy
from logging import getLogger
from threading import Event, Thread

from .__about__ import __version__
from ._vendors.urllib3.exceptions import HTTPError  # type: ignore
from .actions import ActionStore
from .config import CONFIG
from .constants import CHANGELOG_URL, COMPATIBILITY_URL, STATUS_URL
from .deliverer import get_deliverer
from .exceptions import (
    InvalidApplicationName,
    UnsupportedFrameworkVersion,
    UnsupportedPythonVersion,
)
from .frameworks import debug
from .http_client import Urllib3Connection
from .instrumentation import Instrumentation
from .log import configure_root_logger
from .metrics import MetricsStore
from .remote_command import RemoteCommand
from .remote_exception import RemoteException
from .runner import (
    MAX_OBS_QUEUE_LENGTH,
    MAX_QUEUE_LENGTH,
    CappedQueue,
    Runner,
    RunnerSettings,
    RunnerStop,
    process_initial_commands,
)
from .runtime_infos import (
    RuntimeInfos,
    get_parent_cmdline,
    get_process_cmdline,
)
from .session import InvalidSession, InvalidToken, Session
from .signal_client import get_signal_client
from .utils import (
    configure_newrelics_ignore_exception,
    configure_raven_breadcrumbs,
)

RUNNER_THREAD = None
RUNNER_THREAD_PID = None
RUNNER_THREAD_RESTARTING = False
RUNNER_LOCK = threading.RLock()

# Instrumentation timeout, max time waiting for the runner thread to instrument the application
try:
    INSTRUMENTATION_TIMEOUT = int(os.getenv("SQREEN_INSTRUMENTATION_TIMEOUT", 6))
except ValueError:
    INSTRUMENTATION_TIMEOUT = 6

# Graceful timeout, max time waiting for the runner thread to exit before exiting
try:
    GRACEFUL_TIMEOUT = int(os.getenv("SQREEN_GRACEFUL_TIMEOUT", 5))
except ValueError:
    GRACEFUL_TIMEOUT = 5
GRACEFUL_FAIL_MSG = "Sqreen thread didn't exit after %s seconds"

LOGGER = getLogger(__name__)


def uwsgi_version_info():
    """Return uWSGI version tuple if run in uWSGI, None otherwise."""
    # Import must be delayed inside the function, because the uwsgi module is
    # not known by sqreen-start.
    try:
        import uwsgi  # type: ignore
    except ImportError:
        return None
    else:
        return uwsgi.version_info


def should_start_runner_thread():
    """This function contains the logic used to check if the thread
    should be restarted. Beware, this function *must not* call any hooks.
    """
    global RUNNER_THREAD
    global RUNNER_THREAD_PID

    runner_stopped = False
    if RUNNER_THREAD is not None \
            and RUNNER_THREAD.runner is not None \
            and RUNNER_THREAD.runner.stop is True:
        runner_stopped = True

    return RUNNER_THREAD_PID != os.getpid() and not runner_stopped


def before_hook_point():
    global RUNNER_THREAD
    global RUNNER_THREAD_PID
    global RUNNER_THREAD_RESTARTING

    # /!\ The order of execution in this function must not change
    # without a careful review. /!\
    if should_start_runner_thread():
        # It is cheap to first call should_start_runner_thread instead
        # of just locking everytime.
        with RUNNER_LOCK:
            # If we are in a recursive call, do nothing
            if RUNNER_THREAD_RESTARTING is True:
                LOGGER.debug("The runner thread is already restarting")
                return
            # If a concurrent thread has already started the thread, do nothing
            if not should_start_runner_thread():
                LOGGER.info("The runner thread has already been started")
                return
            # Else, restart the thread
            try:
                RUNNER_THREAD_RESTARTING = True
                runner = RUNNER_THREAD.replace()
                runner.start()

                RUNNER_THREAD = runner
                RUNNER_THREAD_PID = os.getpid()

                LOGGER.info(
                    "Sucessfully started runner thread for pid: %d", os.getpid()
                )
            finally:
                RUNNER_THREAD_RESTARTING = False


BLACKLISTED_PARENT_COMMANDS = (
    "ipython",
    "celery worker",
    "rq worker",
    "manage.py shell",
    "pshell",
)
BLACKLISTED_COMMANDS = BLACKLISTED_PARENT_COMMANDS + ("newrelic-admin",)


def get_app_name():
    """ Get the current application name. Return a string if the name is valid.
    Raise InvalidApplicationName otherwise.
    """
    config_app_name = CONFIG["APP_NAME"]
    if config_app_name is not None:
        app_name = config_app_name
    else:
        app_name = "python-application"

    try:
        if isinstance(app_name, bytes):
            app_name = app_name.decode("utf-8")
    except UnicodeDecodeError:
        raise InvalidApplicationName

    # Check the app name is not empty and does not contain
    # any invalid characters (ASCII control chars lower than 0x20)
    if not app_name or any([ord(c) < 0x20 for c in app_name]):
        raise InvalidApplicationName

    return app_name


def get_session(url, token, app_name=None, proxy_url=None):
    """ Create a connection with the endpoint URL and a session. Returns only
    the session.
    """
    LOGGER.warning(
        "Connection to %s%s",
        url,
        " using proxy %s" % proxy_url if proxy_url else "",
    )
    con = Urllib3Connection(url, proxy_url)
    session = Session(con, token, app_name)
    return session


def post_app_sqreen_exception(data):
    """Post a Sqreen exception happening at application level.
    This function does not need a session.

    This function is used to report instrumentation errors. It should never
    fail.
    """
    token = CONFIG["TOKEN"]
    if not token:
        return

    try:
        app_name = get_app_name()
        session = get_session(
            CONFIG["URL"], token, app_name, proxy_url=CONFIG["PROXY_URL"]
        )
        session.post_app_sqreen_exception(data)
    except Exception:
        pass


def start():
    # type: () -> None
    """ Start the background thread and start protection
    """
    # Check if the agent is not disabled first
    try:
        disabled = int(os.getenv("SQREEN_DISABLE", 0))
    except ValueError:
        disabled = 0

    # Configure logging
    configure_root_logger(CONFIG["LOG_LEVEL"], CONFIG["LOG_LOCATION"])

    # Retrieve the command used to launch the process
    command = get_process_cmdline()

    # Retrieve the parent command
    parent_command = get_parent_cmdline()

    # Check if we shouldn't launch ourselves
    for blacklisted_command in BLACKLISTED_COMMANDS:
        if blacklisted_command in command:
            LOGGER.debug(
                "Sqreen agent is disabled when running %s.",
                blacklisted_command,
            )
            return
    for blacklisted_parent_command in BLACKLISTED_PARENT_COMMANDS:
        if blacklisted_parent_command in parent_command:
            LOGGER.debug(
                "Sqreen agent is disabled when running %s.",
                blacklisted_parent_command,
            )
            return
    if hasattr(sys, "argv") and len(sys.argv) >= 2 and sys.argv[1] == "test":
        LOGGER.debug("Sqreen agent is disabled when running tests.")
        return

    if disabled:
        LOGGER.debug("Sqreen agent is disabled.")
        return

    try:
        runtime_infos = RuntimeInfos().all()
    except UnsupportedFrameworkVersion as exception:
        msg = (
            "%s version %s is not supported in this agent version.\n"
            "Sqreen agent is disabled, you're not protected.\n"
            "Documentation can be found at %s.\n"
            "Changelog can be found at %s."
        )
        LOGGER.critical(
            msg,
            exception.framework.title(),
            exception.version,
            COMPATIBILITY_URL,
            CHANGELOG_URL,
        )
        # Alter os.environ?
        return
    except UnsupportedPythonVersion as exception:
        msg = (
            "Python version %s is not supported in this agent version.\n"
            "Sqreen agent is disabled, you're not protected.\n"
            "Documentation can be found at %s.\n"
            "Changelog can be found at %s."
        )
        LOGGER.critical(
            msg, exception.python_version, COMPATIBILITY_URL, CHANGELOG_URL
        )
        return
    except Exception:
        msg = (
            "Runtime information about the system could not be retrieved.\n"
            "Sqreen agent is disabled, you're not protected.\n"
            "Documentation can be found at %s.\n"
        )
        LOGGER.critical(msg, COMPATIBILITY_URL)
        post_app_sqreen_exception(RemoteException.from_exc_info())
        return

    # Configure raven breadcrumbs at start
    configure_raven_breadcrumbs()

    # Ignore Sqreen Exception from APM NewRelics
    configure_newrelics_ignore_exception()

    global RUNNER_THREAD
    global RUNNER_THREAD_PID

    # Check for double instrumentation
    if RUNNER_THREAD is None:
        try:
            runner = RunnerThread(runtime_infos)
            runner.start()

            done = runner.unblock_main_thread.wait(INSTRUMENTATION_TIMEOUT)
            if done is not True:
                post_app_sqreen_exception(RemoteException(
                    "InstrumentationTimeout",
                    "Timeout during instrumentation",
                    exception_payload={
                        "runner_thread_started": runner.started,
                        "runner_thread_was_alive": runner.is_alive(),
                        "instrumentation_enabled": runner.instrumentation.enabled,
                        "runtime_infos": runtime_infos,
                    },
                    stack=runner.stacktrace,
                ).to_dict())

                LOGGER.critical(
                    "Sqreen could not instrument your application in less than %ds.\n"
                    "Please check your account is not expired and your application has network connectivity.\n"
                    "You can also check Sqreen status on %s.",
                    INSTRUMENTATION_TIMEOUT, STATUS_URL)

        except Exception:
            LOGGER.critical(
                "Sqreen thread has failed to start, you're not protected",
                exc_info=True,
            )
            return

        RUNNER_THREAD = runner
        RUNNER_THREAD_PID = os.getpid()


def get_initial_features(config_features, login_features):
    LOGGER.debug("Login features: %s", login_features)

    final_features = copy(login_features)

    if config_features:
        parsed_config_features = {}
        try:
            parsed_config_features = json.loads(config_features)
        except (ValueError, TypeError):
            LOGGER.warning(
                "Invalid config initial features %s",
                config_features,
                exc_info=True,
            )

        if parsed_config_features and isinstance(parsed_config_features, dict):
            msg = "Override login initial features with %s"
            LOGGER.warning(msg, parsed_config_features)

            final_features.update(parsed_config_features)

            LOGGER.warning("Final features %s", final_features)

    return final_features


def get_runner(
    config,
    login_result,
    session,
    queue,
    runtime_infos,
    observations_queue,
    instrumentation,
    settings,
):
    initial_features = get_initial_features(
        config["INITIAL_FEATURES"], login_result.get("features", {})
    )

    # Get a signal client if the feature is enabled
    use_signals = initial_features.get("use_signals", False)
    signal_client = get_signal_client(
        session=session,
        batch_size=initial_features.get("batch_size", 0),
        max_staleness=initial_features.get("max_staleness", 0)
    ) if use_signals else None

    # Get the right deliverer according to initial features
    deliverer = get_deliverer(
        initial_features.get("batch_size", 0),
        initial_features.get("max_staleness", 0),
        session,
        signal_client=signal_client,
    )
    remote_command = RemoteCommand.with_production_commands()
    action_store = ActionStore()
    metrics_store = MetricsStore()
    metrics_store.register_production_aggregators()
    metrics_store.register_default_metrics()

    # Setup the runner
    runner = Runner(
        queue,
        observations_queue,
        session,
        deliverer,
        remote_command,
        runtime_infos,
        instrumentation,
        action_store,
        metrics_store,
        settings,
        initial_features,
        signal_client=signal_client,
    )
    return runner


class RunnerThread(Thread):
    """ Class responsible for starting the runner and monitor it
    """

    name = "SqreenRunnerThread"

    def __init__(self, runtime_infos, queue=None, observations_queue=None,
                 instrumentation=None, settings=None):
        super(RunnerThread, self).__init__()
        self.daemon = True
        self.queue = queue or CappedQueue(MAX_QUEUE_LENGTH)
        self.observations_queue = observations_queue or CappedQueue(MAX_OBS_QUEUE_LENGTH)
        self.unblock_main_thread = Event()
        self.runtime_infos = runtime_infos
        self.instrumentation = instrumentation or Instrumentation(
            self.observations_queue, self.queue, before_hook_point
        )
        self.settings = settings or RunnerSettings()
        self.runner = None
        self.started = False

    @property
    def stacktrace(self):
        """Get the runner thread stack trace if it exists.
        """
        frame = sys._current_frames().get(self.ident, None)
        if frame is not None:
            return traceback.extract_stack(frame)

    def run(self):
        """ Launch the runner
        """
        self.started = True
        LOGGER.debug("Starting Sqreen %s", __version__)
        # FIXME this part should be more tested
        while True:
            session = None
            self.runner = None

            try:
                token = CONFIG["TOKEN"]
                if not token:
                    msg = (
                        "Sorry but we couldn't find your Sqreen token.\n"
                        "Your application is NOT currently protected by Sqreen.\n"
                        "\n"
                        "Have you filled your sqreen.ini?"
                    )
                    LOGGER.critical(msg)
                    self.unblock_main_thread.set()
                    return

                try:
                    app_name = get_app_name()
                except InvalidApplicationName:
                    msg = (
                        "Sorry but your application name seems invalid.\n"
                        "Your application is NOT currently protected by Sqreen.\n"
                        "\n"
                        "Please visit the following link: https://docs.sqreen.com/faq/org-token-migration/#application-name-limitations"
                    )
                    LOGGER.critical(msg)
                    self.unblock_main_thread.set()
                    return

                LOGGER.warning("Using token %r and application name %r", token, app_name)

                # Initiate HTTP connection to the backend and the session
                session = get_session(
                    CONFIG["URL"], token, app_name, proxy_url=CONFIG["PROXY_URL"]
                )

                runtime_infos = dict(self.runtime_infos)
                runtime_infos["various_infos"] = {
                    k: v
                    for k, v in runtime_infos.get("various_infos", {}).items()
                    if k != "dependencies"
                }
                try:
                    login_result = session.login(runtime_infos)
                except InvalidToken:
                    msg = (
                        "Sorry but your Sqreen token appears to be invalid.\n"
                        "Your application is NOT currently protected by Sqreen.\n"
                        "\n"
                        "Please check the token against the interface."
                    )
                    LOGGER.critical(msg)
                    self.unblock_main_thread.set()
                    return
                except HTTPError:
                    msg = (
                        "Sorry but Sqreen's backend appears to be unavailable.\n"
                        "Your application is NOT currently protected by Sqreen.\n"
                        "\n"
                        "Please check http://sqreen.status.io for outage.\n"
                        "\n"
                        "Otherwise, back.sqreen.io may be blocked by your server's network.\n"
                        "If necessary, you may tell Sqreen to connect over a proxy.\n"
                        "Please visit https://docs.sqreen.com/python/configuration/ for more details."
                    )
                    LOGGER.critical(msg)
                    self.unblock_main_thread.set()
                    return

                LOGGER.info("Login success")

                self.runner = get_runner(
                    CONFIG,
                    login_result,
                    session,
                    self.queue,
                    self.runtime_infos,
                    self.observations_queue,
                    self.instrumentation,
                    self.settings,
                )

                # Process the initial commands
                process_initial_commands(login_result, self.runner)

                retry = 3
                while True:
                    try:
                        # if we did not receive instrumentation_enable, try
                        # to get it faster before the MainThread stops waiting.
                        while not self.instrumentation.enabled and retry > 0:
                            retry -= 1
                            self._random_sleep(2.0)
                            self.runner.do_heartbeat()
                        self.unblock_main_thread.set()
                        self.runner.run()
                    except HTTPError:
                        LOGGER.exception(
                            "An HTTP error occurred, restarting runner"
                        )
                        try:
                            session.post_sqreen_exception(
                                RemoteException.from_exc_info().to_dict()
                            )
                        except HTTPError:
                            pass
                        self._random_sleep()
                    else:
                        # If the runner exits normally, return.
                        return
            except InvalidSession:
                LOGGER.info("Relaunching the session because it was dropped by the backend.")
                self._random_sleep()
            except Exception as exc:
                LOGGER.exception("An unexpected exception occurred")

                callback_payload = getattr(exc, "callback_payload", None)
                exception_payload = getattr(exc, "exception_payload", None)
                request_payload = getattr(exc, "request_payload", None)
                remote_exception = RemoteException.from_exc_info(
                    callback_payload=callback_payload,
                    exception_payload=exception_payload,
                    request_payload=request_payload,
                )

                if session is not None and session.is_connected():
                    try:
                        session.post_sqreen_exception(
                            remote_exception.to_dict()
                        )
                        session.logout()
                    except HTTPError:
                        LOGGER.exception("An HTTP error occurred while logout")
                    except Exception:
                        LOGGER.exception("Exception while logout")
                        return
                else:
                    post_app_sqreen_exception(remote_exception.to_dict())

                try:
                    self.instrumentation.deinstrument_all()
                except Exception:
                    # We did not managed to remove instrumentation, state is unclear:
                    # terminate thread
                    LOGGER.exception("Exception while trying to clean-up")
                    return

                self._random_sleep()

    def quick_exit(self):
        """Check for early exit."""
        if os.environ.get("SQREEN_SKIP_LOGOUT", False):
            return True
        # uWSGI 2.0.15 may segfault upon exiting.
        # See https://github.com/unbit/uwsgi/issues/1651.
        uwsgi_vers = uwsgi_version_info()
        if uwsgi_vers and uwsgi_vers >= (2, 0, 15):
            return True
        return self.settings.get_debug_flag() or debug.detect_debug_mode()

    def start(self):
        """Start the runner thread."""
        atexit.register(self.graceful_stop)
        return super(RunnerThread, self).start()

    def graceful_stop(self, force_quick_exit=False):
        """Gracefully stop the runner thread and wait for its termination."""
        self.queue.put(RunnerStop)

        if force_quick_exit or self.quick_exit():
            LOGGER.warning("Quick exit, don't logout")
            return

        try:
            self.join(GRACEFUL_TIMEOUT)
        except RuntimeError:
            logger = getLogger(__name__)
            logger.warning(GRACEFUL_FAIL_MSG, GRACEFUL_TIMEOUT)
        else:
            if self.isAlive():
                logger = getLogger(__name__)
                logger.warning(GRACEFUL_FAIL_MSG, GRACEFUL_TIMEOUT)

    def replace(self, **kwargs):
        """Create a new runner thread instance."""
        nkwargs = dict(
            queue=self.queue,
            observations_queue=self.observations_queue,
            instrumentation=self.instrumentation,
            settings=self.settings,
        )
        nkwargs.update(kwargs)
        return self.__class__(RuntimeInfos().all(), **nkwargs)

    @classmethod
    def _random_sleep(cls, max_delay=10.0):
        """Sleep for a random duration before trying again."""
        delay = random.uniform(0, max_delay)
        LOGGER.info("Sleeping %0.2fs before retry", delay)
        time.sleep(delay)



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#

__author__ = "Sqreen"
__copyright__ = "Copyright 2016 - 2020, Sqreen"
__email__ = "contact@sqreen.io"
__license__ = "proprietary"
__version__ = "1.21.0b4"



# -*- coding: utf-8 -*-
# Copyright (c) 2016 - 2020 Sqreen. All rights reserved.
# Please refer to our terms for more information:
#
#     https://www.sqreen.io/terms.html
#
""" Aggregate authentication tentatives
"""
import json
import logging

from ..rules import RuleCallback
from ..utils import CustomJSONEncoder
from .matcher_callback import MatcherRule

LOGGER = logging.getLogger(__name__)


class AuthMetricsMixin:

    def record_auth_metrics(self, auth_status, identifiers):
        request = self.storage.get_current_request()
        if not request:
            LOGGER.debug("No request was recorded, abort")
            return

        key = {
            "keys": [(k.lower(), v) for k, v in identifiers.items()],
            "ip": request.client_ip
        }
        observation_key = json.dumps(
            key, separators=(",", ":"), sort_keys=True, cls=CustomJSONEncoder
        )
        self.record_observation(auth_status, observation_key, 1)


class DjangoAuthMetrics(AuthMetricsMixin, MatcherRule):

    INTERRUPTIBLE = False

    def post(self, instance, args, kwargs, result=None, **options):
        # Django authentication model return either an User or None
        auth_status = "auto-login-fail" if result is None else "auto-login-success"

        # Search for credentials identifier that match the whitelist
        identifiers = {k: v for k, v in kwargs.items()
                       if self.match(k) and v is not None}
        if identifiers:
            self.record_auth_metrics(auth_status, identifiers)


# Compatibility callback name
AuthMetricsCB = DjangoAuthMetrics


class SDKAuthMetrics(AuthMetricsMixin, RuleCallback):

    INTERRUPTIBLE = False

    def pre(self, instance, args, kwargs, **options):
        """Report a metric for SDK authentication tracking."""
        if args and kwargs:
            auth_status = "sdk-login-success" if args[0] else "sdk-login-fail"
            self.record_auth_metrics(auth_status, kwargs)


class SDKSignupMetrics(AuthMetricsMixin, RuleCallback):

    INTERRUPTIBLE = False

    def pre(self, instance, args, kwargs, **options):
        """Report a metric for SDK signup tracking."""
        if kwargs:
            self.record_auth_metrics("sdk-signup", kwargs)
